import pandas
import numpy
import copy

from .controller import Model5c as _Model5c
from ..dataframes import DataFrames, get_dataframe_format
from .linear import LinearFunction_C, DictOfLinearFunction_C
from ..general_precision import l4_float_dtype
from .constraints import ParametricConstraintList
from typing import Sequence, Mapping
from numbers import Number
import numpy as np



import logging
from ..log import logger_name
logger = logging.getLogger(logger_name+'.model')

class Model(_Model5c):
	"""A discrete choice model.

	Parameters
	----------
	parameters : Sequence, optional
		The names of parameters used in this model.  It is generally not
		necessary to define parameter names at initialization, as the names
		can (and will) be collected from the utility function and nesting
		components later.
	utility_ca : LinearFunction_C, optional
		The utility_ca function, which represents the qualitative portion of
		utility for attributes that vary by alternative.
	utility_co : DictOfLinearFunction, optional
		The utility_co function, which represents the qualitative portion of
		utility for attributes that vary by decision maker but not by alternative.
	quantity_ca : LinearFunction_C, optional
		The quantity_ca function, which represents the quantitative portion of
		utility for attributes that vary by alternative.
	quantity_scale : str, optional
		The name of the parameter used to scale the quantitative portion of
		utility.
	graph : NestingTree, optional
		The nesting tree for this choice model.
	dataservice : DataService, optional
		An object that can act as a DataService to generate the data needed for
		this model.

	"""

	utility_co = DictOfLinearFunction_C()
	"""DictOfLinearFunction_C : The portion of the utility function computed from idco data.
	
	The keys of this mapping are alternative codes for the applicable elemental
	alteratives, and the values are linear functions to compute for the indicated
	alternative.  Each alternative that has any idco utility components must have
	a unique linear function given.
	"""

	utility_ca = LinearFunction_C()
	"""LinearFunction_C : The portion of the utility function computed from idca data.
	
	Examples
	--------
		
	>>> from larch import Model, P, X
	>>> m = Model()
	>>> m.utility_ca = P.Param1 * X.Data1 + P.Param2 * X.Data2
	>>> print(m.utility_ca)
	P.Param1 * X.Data1 + P.Param2 * X.Data2
	
	"""

	quantity_ca = LinearFunction_C()
	"""LinearFunction_C : The portion of the quantity function computed from idca data.
	
	Note that for the quantity function, the actual computed linear function
	uses the exponential of the parameter value(s), not the raw values. Thus, 
	if the quantity function is given as `P.Param1 * X.Data1 + P.Param2 * X.Data2`,
	the computed values will actually be `exp(P.Param1) * X.Data1 + exp(P.Param2) * X.Data2`.
	This transformation ensures that the outcome from the quantity function is 
	always positive, so long as at all of the data terms in the function are
	positive.  The `LinearFunction_C` class itself is not intrinsically aware
	of this implementation detail, but the `Model.utility_functions()` method is, 
	and will render the complete utility function in a mathematically correct form.
	
	Examples
	--------
		
	>>> from larch import Model, P, X
	>>> m = Model()
	>>> m.quantity_ca = P.Param1 * X.Data1 + P.Param2 * X.Data2
	>>> print(m.quantity_ca)
	P.Param1 * X.Data1 + P.Param2 * X.Data2

	"""

	constraints = ParametricConstraintList()

	@classmethod
	def Example(cls, n=1):
		from ..examples import example
		return example(n)

	@classmethod
	def load(cls, filename):
		import os
		if not os.path.exists(filename):
			raise FileNotFoundError(filename)
		_, ext = os.path.splitext(filename)
		if ext == '.xlsx':
			from ..util.excel import load_metadata
			return load_metadata(filename, key='_self_')
		import pickle
		return pickle.load(filename)

	def __init__(self,
				 utility_ca=None,
				 utility_co=None,
				 quantity_ca=None,
				 constraints=None,
				 **kwargs):
		import sys
		self._sklearn_data_format = 'idce'
		self.utility_co = utility_co
		self.utility_ca = utility_ca
		self.quantity_ca = quantity_ca
		self.constraints = constraints
		super().__init__(**kwargs)
		self._scan_all_ensure_names()
		self.mangle()

	def dumps(self):
		"""
		Use pickle to dump the contents of this Model to a bytestring.

		Any associated data (dataframes and dataservice) are not included.

		Returns
		-------
		bytes
		"""
		import pickle
		return pickle.dumps(self)



	def get_params(self, deep=True):
		p = dict()
		if deep:
			p['frame'] = self.pf.copy()
			p['utility_ca'] = LinearFunction_C(self.utility_ca.copy())
			p['utility_co'] = self.utility_co.copy_without_touch_callback()
			p['quantity_ca'] = LinearFunction_C(self.quantity_ca.copy())
			p['quantity_scale'] = self.quantity_scale.copy() if self.quantity_scale is not None else None
			p['graph'] = copy.deepcopy(self.graph)
			p['constraints'] = copy.deepcopy(self.constraints)
			p['is_clone'] = True
		else:
			p['frame'] = self.pf
			p['utility_ca'] = self.utility_ca
			p['utility_co'] = self.utility_co
			p['quantity_ca'] = self.quantity_ca
			p['quantity_scale'] = self.quantity_scale
			p['graph'] = self.graph
			p['constraints'] = self.constraints
			p['is_clone'] = True
		return p

	def set_params(self, **kwargs):
		if 'frame' in kwargs and kwargs['frame'] is not None:
			self._frame = kwargs['frame']

		self.utility_ca = kwargs.get('utility_ca', None)
		self.utility_co = kwargs.get('utility_co', None)
		self.quantity_ca = kwargs.get('quantity_ca', None)
		self.quantity_scale = kwargs.get('quantity_scale', None)
		self.graph = kwargs.get('graph', None)
		self.constraints = kwargs.get('constraints', [])


	def fit(self, X, y, sample_weight=None, **kwargs):
		"""Estimate the parameters of this model from the training set (X, y).

		Parameters
		----------
		X : pandas.DataFrame
			This DataFrame can be in idca, idce, or idco formats.
			If given in idce format, this is a DataFrame with *n_casealts* rows, and
			a two-level MultiIndex.
		y : array-like or str
			The target choice values.  If given as a ``str``, use that named column of `X`.
		sample_weight : array-like, shape = [n_cases] or [n_casealts], or None
			Sample weights. If None, then samples are equally weighted. If shape is *n_casealts*,
			the array is collapsed to *n_cases* by taking only the first weight in each case.

		Returns
		-------
		self : Model
		"""

		if not isinstance(X, pandas.DataFrame):
			raise TypeError(f'must fit on an {self._sklearn_data_format} dataframe')

		if sample_weight is not None:
			raise NotImplementedError('sample_weight not implemented')

		self._sklearn_data_format = get_dataframe_format(X)

		if self._sklearn_data_format == 'idce':

			if sample_weight is not None:
				if isinstance(sample_weight, str):
					sample_weight = X[sample_weight]
				if len(sample_weight) == X.shape[0]:
					sample_weight = sample_weight.groupby(X.index.codes[0]).first()

			if isinstance(y, str):
				y = X[y].unstack().fillna(0)
			elif isinstance(y, (pandas.DataFrame, pandas.Series)):
				y = y.unstack().fillna(0)
			else:
				y = pandas.Series(y, index=X.index).unstack().fillna(0)

			from ..dataframes import _check_dataframe_of_dtype
			try:
				if _check_dataframe_of_dtype(X, l4_float_dtype):
					# when the dataframe is an array of the correct type,
					# it is efficient to use it directly
					self.dataframes = DataFrames(
						ce = X,
						ch = y,
						wt = sample_weight,
					)
				else:
					# when the dataframe is not an array of the correct type,
					# it is efficient to only manipulate needed columns
					self.dataframes = DataFrames(
						ce = X[self.required_data().ca],
						ch = y,
						wt = sample_weight,
					)
			except KeyError:
				# not all keys were available in natural form, try computing them
				dfs1 = DataFrames( ce = X, )
				dfs = dfs1.make_dataframes(self.required_data())
				dfs.data_ch = y
				dfs.data_wt = sample_weight
				self.dataframes = dfs
		else:
			raise NotImplementedError(self._sklearn_data_format)

		self.maximize_loglike(**kwargs)

		return self

	def predict(self, X):
		"""Predict choices for X.

		This method returns the index of the maximum probability choice, not the probability.
		To recover the probability, which is probably what you want (pun intended), see
		:meth:`predict_proba`.

		Parameters
		----------
		X : pandas.DataFrame

		Returns
		-------
		y : array of shape = [n_cases]
			The predicted choices.
		"""
		if not isinstance(X, pandas.DataFrame):
			raise TypeError("X must be a pandas.DataFrame")

		if self._sklearn_data_format in ('idce', 'idca'):
			pr = self.predict_proba(X)
			pr = pr.unstack()
		elif self._sklearn_data_format in ('idco',):
			pr = self.predict_proba(X)
		else:
			raise NotImplementedError(self._sklearn_data_format)

		result = numpy.nanargmax(pr.values, axis=1)

		if self._sklearn_data_format in ('idce', 'idca'):
			pr.values[~numpy.isnan(pr.values)] = 0
			pr.values[numpy.arange(pr.shape[0]), result] = 1
			result = pr.stack()

		return result

	def predict_proba(self, X):
		"""Predict probability for X.

		Parameters
		----------
		X : pandas.DataFrame

		Returns
		-------
		y : array of shape = [n_cases, n_alts]
			The predicted probabilities.
		"""

		if not isinstance(X, pandas.DataFrame):
			raise TypeError(f'predict_proba requires an {self._sklearn_data_format} dataframe')

		if self._sklearn_data_format == 'idce':
			try:
				self.dataframes = DataFrames(
					ce = X[self.required_data().ca],
				)
			except KeyError:
				# not all keys were available in natural form, try computing them
				dfs1 = DataFrames( ce = X, )
				dfs = dfs1.make_dataframes({'ca':self.required_data().ca})
				self.dataframes = dfs

		elif self._sklearn_data_format == 'idca':
			try:
				self.dataframes = DataFrames(
					ca = X[self.required_data().ca],
				)
			except KeyError:
				# not all keys were available in natural form, try computing them
				dfs1 = DataFrames( ca = X, )
				dfs = dfs1.make_dataframes({'ca':self.required_data().ca})
				self.dataframes = dfs
		elif self._sklearn_data_format == 'idco':
			try:
				self.dataframes = DataFrames(
					co = X[self.required_data().co],
				)
			except KeyError:
				# not all keys were available in natural form, try computing them
				dfs1 = DataFrames(co=X, )
				dfs = dfs1.make_dataframes({'co': self.required_data().co})
				self.dataframes = dfs
		else:
			raise NotImplementedError(self._sklearn_data_format)

		result = self.probability(return_dataframe=self._sklearn_data_format)

		return result

	def score(self, X, y, sample_weight=None):
		"""
		Returns the mean negative log loss on the given test data and labels.

		Note that the log loss is defined as the negative of the log likelihood,
		and thus the mean negative log loss is also just mean log likelihood.

		Parameters
		----------
		X : pandas.DataFrame
			If given in idce format, a dataFrame with *n_casealts* rows.
		y : array-like or str
			The target choice values.  If given as a ``str``, use that named column of `X`.
		sample_weight : array-like, shape = [n_casealts], or None
			Sample weights. If None, then samples are equally weighted.

		Returns
		-------
		score : float
			Mean negative log loss of self.predict_proba(X) wrt. y.
		"""
		if isinstance(y, str):
			y = X[y].values.reshape(-1)
		elif isinstance(y, (pandas.DataFrame, pandas.Series)):
			y = y.values.reshape(-1)
		else:
			y = y.reshape(-1)

		pr = self.predict_proba(X)

		weight_adjust = numpy.sum(y) / self.dataframes.n_cases

		if sample_weight is not None:
			sample_weight = sample_weight[y>0] / weight_adjust

		pr = pr[y>0]
		y = y[y>0]

		if sample_weight is None:
			return numpy.sum(numpy.log(pr) * y / weight_adjust) / self.dataframes.n_cases
		else:
			return numpy.sum(numpy.log(pr) * y * sample_weight) / numpy.sum(sample_weight)





	def __repr__(self):
		s = "<larch.Model"
		if self.is_mnl():
			s += " (MNL)"
		else:
			s += " (GEV)"
		if self.title != "Untitled":
			s += f' "{self.title}"'
		s += ">"
		return s


	def utility_functions(self, subset=None, resolve_parameters=False):
		"""
		Generate an XHTML output of the utility function(s).

		Parameters
		----------
		subset : Collection, optional
			A collection of alternative codes to include. This only has effect if
			there are separate utility_co functions set by alternative. It is
			recommended to use this parameter if there are a very large number of
			alternatives, and the utility functions of most (or all) of them
			can be effectively communicated by showing only a few.
		resolve_parameters : bool, default False
			Whether to resolve the parameters to the current (estimated) value
			in the output.

		Returns
		-------
		xmle.Elem
		"""
		self.unmangle()
		from xmle import Elem
		x = Elem('div')
		t = x.elem('table', style="margin-top:1px;", attrib={'class':'floatinghead'})
		if len(self.utility_co):
			# t.elem('caption', text=f"Utility Functions",
			# 	   style="caption-side:top;text-align:left;font-family:Roboto;font-weight:700;"
			# 			 "font-style:normal;font-size:100%;padding:0px;color:black;")

			# iterate over all alternatives if a dataframes is attached and lists the alternatives
			try:
				if self.dataservice is not None:
					alts = self.dataservice.alternative_codes()
				elif self.dataframes is not None:
					alts = self.dataframes.alternative_codes()
				else:
					alts = self.utility_co.keys()
			except:
				alts = self.utility_co.keys()
			t_head = t.elem('thead')
			tr = t_head.elem('tr')
			tr.elem('th', text="alt")
			tr.elem('th', text='formula', attrib={'style':'text-align:left;'})
			t_body = t.elem('tbody')
			for j in alts:
				if subset is None or j in subset:
					tr = t_body.elem('tr')
					tr.elem('td', text=str(j))
					utilitycell = tr.elem('td', attrib={'style':'text-align:left;'})
					utilitycell.elem('div')
					anything = False
					if len(self.utility_ca):
						utilitycell[-1].tail = (utilitycell[-1].tail or "") + " + "
						utilitycell << list(self.utility_ca.__xml__(linebreaks=True, resolve_parameters=self, value_in_tooltips=not resolve_parameters))
						anything = True
					if j in self.utility_co:
						v = self.utility_co[j]
						if len(v):
							if anything:
								utilitycell << Elem('br')
							utilitycell[-1].tail = (utilitycell[-1].tail or "") + " + "
							utilitycell << list(v.__xml__(linebreaks=True, resolve_parameters=self, value_in_tooltips=not resolve_parameters))
							anything = True
					if len(self.quantity_ca):
						if anything:
							utilitycell << Elem('br')
						if self.quantity_scale:
							utilitycell[-1].tail = (utilitycell[-1].tail or "") + " + "
							from .linear import ParameterRef_C
							utilitycell << list(ParameterRef_C(self.quantity_scale).__xml__(resolve_parameters=self, value_in_tooltips=not resolve_parameters))
							utilitycell[-1].tail = (utilitycell[-1].tail or "") + " * log("
						else:
							utilitycell[-1].tail = (utilitycell[-1].tail or "") + " + log("
						content = self.quantity_ca.__xml__(linebreaks=True, lineprefix="  ",
														   exponentiate_parameters=True, resolve_parameters=self, value_in_tooltips=not resolve_parameters)
						utilitycell << list(content)
						utilitycell.elem('br', tail=")")
		else:
			# there is no differentiation by alternatives, just give one formula
			# t.elem('caption', text=f"Utility Function",
			# 	   style="caption-side:top;text-align:left;font-family:Roboto;font-weight:700;"
			# 			 "font-style:normal;font-size:100%;padding:0px;color:black;")
			tr = t.elem('tr')
			utilitycell = tr.elem('td', attrib={'style':'text-align:left;'})
			utilitycell.elem('div')
			anything = False
			if len(self.utility_ca):
				utilitycell[-1].tail = (utilitycell[-1].tail or "") + " + "
				utilitycell << list(self.utility_ca.__xml__(linebreaks=True, resolve_parameters=self, value_in_tooltips=not resolve_parameters))
				anything = True
			if len(self.quantity_ca):
				if anything:
					utilitycell << Elem('br')
				if self.quantity_scale:
					utilitycell[-1].tail = (utilitycell[-1].tail or "") + " + "
					from .linear import ParameterRef_C
					utilitycell << list(ParameterRef_C(self.quantity_scale).__xml__(resolve_parameters=self, value_in_tooltips=not resolve_parameters))
					utilitycell[-1].tail = (utilitycell[-1].tail or "") + " * log("
				else:
					utilitycell[-1].tail = (utilitycell[-1].tail or "") + " + log("
				content = self.quantity_ca.__xml__(linebreaks=True, lineprefix="  ", exponentiate_parameters=True, resolve_parameters=self, value_in_tooltips=not resolve_parameters)
				utilitycell << list(content)
				utilitycell.elem('br', tail=")")
		return x

	def _utility_functions_as_frame(self, subset=None, resolve_parameters=False):
		"""
		Generate a tabular output of the utility function(s).

		Parameters
		----------
		subset : Collection, optional
			A collection of alternative codes to include. This only has effect if
			there are separate utility_co functions set by alternative. It is
			recommended to use this parameter if there are a very large number of
			alternatives, and the utility functions of most (or all) of them
			can be effectively communicated by showing only a few.
		resolve_parameters : bool, default False
			Whether to resolve the parameters to the current (estimated) value
			in the output.  Not implemented.

		Returns
		-------
		xmle.Elem
		"""
		self.unmangle()

		tf = pandas.DataFrame(
			index=pandas.MultiIndex.from_tuples([], names=['Alt', 'Line']),
			columns=['Formula'],
		)

		if len(self.utility_co):
			# iterate over all alternatives if a dataframes is attached and lists the alternatives
			try:
				if self.dataservice is not None:
					alts = self.dataservice.alternative_codes()
				elif self.dataframes is not None:
					alts = self.dataframes.alternative_codes()
				else:
					alts = self.utility_co.keys()
			except:
				alts = self.utility_co.keys()

			for a in alts:
				if not (subset is None or a in subset):
					continue
				line = 1
				op = ' '
				for i in self.utility_ca:
					tf.loc[(a, line), :] = f"{op} {str(i)}"
					op = '+'
					line += 1
				if a in self.utility_co:
					for i in self.utility_co[a]:
						tf.loc[(a, line), :] = f"{op} {str(i)}"
						op = '+'
						line += 1
				if len(self.quantity_ca):
					if self.quantity_scale:
						from .linear import ParameterRef_C
						q = ParameterRef_C(self.quantity_scale)
						tf.loc[(a, line), :] = f"{op} {str(q)} * log("
					else:
						tf.loc[(a, line), :] = f"{op} log("
					op = ' '
					line += 1
					for i in self.quantity_ca:
						tf.loc[(a, line), :] = f"    {op} {i._str_exponentiate()}"
						op = '+'
						line += 1
					tf.loc[(a, line), :] = f")"
					op = '+'
					line += 1

		else:
			# there is no differentiation by alternatives, just give one formula
			a = '*'
			line = 1
			op = ' '
			for i in self.utility_ca:
				tf.loc[(a, line), :] = f"{op} {str(i)}"
				op = '+'
				line += 1
			if len(self.quantity_ca):
				if self.quantity_scale:
					from .linear import ParameterRef_C
					q = ParameterRef_C(self.quantity_scale)
					tf.loc[(a, line), :] = f"{op} {str(q)} * log("
				else:
					tf.loc[(a, line), :] = f"{op} log("
				op = ' '
				line += 1
				for i in self.quantity_ca:
					tf.loc[(a, line), :] = f"    {op} {i._str_exponentiate()}"
					op = '+'
					line += 1
				tf.loc[(a, line), :] = f")"
				op = '+'
				line += 1

		return tf

	def required_data(self):
		"""
		What data is required in DataFrames for this model to be used.

		Returns
		-------
		dictx
		"""
		try:
			from ..util import dictx
			req_data = dictx()

			if self.utility_ca is not None and len(self.utility_ca):
				if 'ca' not in req_data:
					req_data.ca = set()
				for i in self.utility_ca:
					req_data.ca.add(str(i.data))

			if self.quantity_ca is not None and len(self.quantity_ca):
				if 'ca' not in req_data:
					req_data.ca = set()
				for i in self.quantity_ca:
					req_data.ca.add(str(i.data))

			if self.utility_co is not None and len(self.utility_co):
				if 'co' not in req_data:
					req_data.co = set()
				for alt, func in self.utility_co.items():
					for i in func:
						if str(i.data)!= '1':
							req_data.co.add(str(i.data))

			if 'ca' in req_data:
				req_data.ca = list(sorted(req_data.ca))
			if 'co' in req_data:
				req_data.co = list(sorted(req_data.co))

			if self.choice_ca_var:
				req_data.choice_ca = self.choice_ca_var
			elif self.choice_co_vars:
				req_data.choice_co = self.choice_co_vars
			elif self.choice_co_code:
				req_data.choice_co_code = self.choice_co_code
			elif self.choice_any:
				req_data.choice_any = True

			if self.weight_co_var:
				req_data.weight_co = self.weight_co_var

			if self.availability_var:
				req_data.avail_ca = self.availability_var
			elif self.availability_co_vars:
				req_data.avail_co = self.availability_co_vars
			elif self.availability_any:
				req_data.avail_any = True

			return req_data
		except:
			logger.exception("error in required_data")

	def __contains__(self, item):
		return (item in self.pf.index) #or (item in self.rename_parameters)

	def doctor(
			self,
			repair_ch_av=None,
			repair_ch_zq=None,
			repair_asc=None,
			repair_noch_nowt=None,
			repair_nan_wt=None,
			verbose=3,
	):
		self.unmangle(True)
		from ..troubleshooting import doctor
		return doctor(
			self,
			repair_ch_av=repair_ch_av,
			repair_ch_zq=repair_ch_zq,
			repair_asc=repair_asc,
			repair_noch_nowt=repair_noch_nowt,
			repair_nan_wt=repair_nan_wt,
			verbose=verbose,
		)

	def copy(self):
		"""
		Create a copy of the Model.

		The copy shares the same DataService, but does not
		share or copy any loaded DataFrames, which must be
		reloaded if desired.

		Returns
		-------
		Model
		"""
		import pickle
		result = pickle.loads(self.dumps())
		result.dataservice = self.dataservice
		result.graph.set_touch_callback(result.mangle)
		return result

	def remove_unused_parameters(self):
		"""
		Remove parameters that are not used in the model.
		"""
		self.unmangle(True)
		old_pf = self.pf.copy()
		self.pf.drop(self.pf.index, inplace=True)
		self.unmangle(True)
		overlay = pandas.merge(self.pf.iloc[:, 0:0], old_pf, left_index=True, right_index=True, how='left')
		for col in overlay.columns:
			f = ~overlay[col].isna()
			self.pf.loc[f, col] = overlay.loc[f, col]



	def data_statistics(self):
		"""
		Generate a summary of data statistics.

		This summary includes several sections, including
		statistics on idco and idca data as appropriate, as well as
		choice and availability statistics.

		Returns
		-------
		xmle.Elem
		"""
		if self.dataframes is not None:
			return self.dataframes.statistics(graph=self.graph)

	def _get_constraints(self, method):
		if method.lower() in ('slsqp', 'cobyla'):
			constraint_dicts = []
			for c in self.constraints:
				constraint_dicts.extend(c.as_constraint_dicts())
			return constraint_dicts
		if method.lower() in ('trust-constr'):
			constraints = []
			for c in self.constraints:
				constraints.extend(c.as_linear_constraints())
			return constraints
		return ()

	def _add_constraint(self, c):
		"""
		Add a parametric constraint to the model.

		Parameters
		----------
		c : ParametricConstraint
			The constraint to potentially add.  Constraints are
			not actually added if they duplicate an existing
			constraint.

		Returns
		-------
		bool
			Whether this constraint was actually added to the model.
		"""
		from .constraints import ParametricConstraint
		if not isinstance(c, ParametricConstraint):
			raise TypeError(f"constraint must be ParametricConstraint not {type(c)}")
		duplicate_constraint = False
		for i in self.constraints:
			if i==c:
				duplicate_constraint = True
				break
		if not duplicate_constraint:
			self.constraints.append(c)

	def _get_bounds_constraints(self, binding_tol=1e-4):
		"""
		Convert bounds to parametric constraints on the model.

		Parameters
		----------
		binding_tol : Number or Mapping
			The binding tolerance to use, which determines
			whether a constraint is considered active or not.

		Returns
		-------
		list
			A list of constraints.
		"""
		from .constraints import FixedBound
		if isinstance(binding_tol, Number):
			default_binding_tol = binding_tol
		else:
			default_binding_tol = 1e-4

		if not isinstance(binding_tol, Mapping):
			binding_tol = {}

		get_bind_tol = lambda x: binding_tol.get(x, default_binding_tol)

		bounds = []
		self.unmangle()
		for pname, pf_row in self.pf.iterrows():
			if pf_row['holdfast']:
				# don't create bounds constraints on holdfast parameters
				continue
			b = (pf_row['minimum'], pf_row['maximum'])
			if b[0] != -numpy.inf or b[1] != numpy.inf:
				bounds.append(FixedBound(pname, *b, model=self, binding_tol=get_bind_tol(pname)))

		return bounds

	def get_all_contraints(self):
		"""
		All constraints, including bounds.

		Returns
		-------
		list
		"""
		constraints = list(self.constraints)
		try:
			constraints.extend(self._get_bounds_constraints())
		except:
			pass
		return constraints


	def get_all_binding_contraints(self):
		"""
		All binding constraints, including min-max bounds.

		Returns
		-------
		list
		"""
		binding_constraints = list()

		for c in self.get_all_contraints():
			if np.absolute(c.fun(self.pf.value)) < c.binding_tol:
				binding_constraints.append(c)

		return binding_constraints






from .abstract_model import AbstractChoiceModel
from . import persist_flags
from ..exceptions import ParameterNotInModelWarning
from .constraints import ParametricConstraintList
from collections.abc import MutableSequence

class ModelGroup(AbstractChoiceModel, MutableSequence):

	constraints = ParametricConstraintList()

	def __init__(
			self,
			models,
			*,
			parameters=None,
			frame=None,
			title=None,
			dataservice=None,
			constraints=None,
	):
		super().__init__(
			parameters=parameters,
			frame=frame,
			title=title,
		)
		self._k_models = list(models)
		self._dataservice = dataservice
		self._dataframes = None
		self._mangled = True
		self.constraints = constraints

	def __getitem__(self, x):
		return self._k_models[x]

	def __setitem__(self, i, value):
		assert isinstance(value, AbstractChoiceModel)
		self._k_models[i] = value

	def __delitem__(self, x):
		del self._k_models[x]

	def __len__(self):
		return len(self._k_models)

	def __contains__(self, x):
		return (x in self.pf.index)

	def insert(self, i, value):
		assert isinstance(value, AbstractChoiceModel)
		self._k_models.insert(i,value)

	@property
	def dataframes(self):
		return 'access dataframes of group members individually'

	@property
	def n_cases(self):
		"""int : Total number of cases in the attached dataframes of all grouped models."""
		return sum(k.n_cases for k in self._k_models)

	def total_weight(self):
		"""
		The total weight of cases in the attached dataframes of all grouped models.

		Returns
		-------
		float
		"""
		return sum(k.total_weight() for k in self._k_models)

	def unmangle(self):
		super().unmangle()
		for k in self._k_models:
			for i in k.pf.index:
				if i not in self._frame.index:
					self._frame.loc[i] = k.pf.loc[i]

	def _frame_values_have_changed(self):
		for k in self._k_models:
			k._frame_values_have_changed()

	def __prep_for_compute(self, x=None):
		self.unmangle()
		if x is not None:
			self.set_values(x)
		vals = self.pf.value
		import warnings
		with warnings.catch_warnings():
			warnings.simplefilter("ignore", category=ParameterNotInModelWarning)
			for k in self._k_models:
				k.set_values(**vals)

	def load_data(self, dataservice=None, autoscale_weights=True, log_warnings=True):
		for k in self._k_models:
			k.load_data(
				dataservice=dataservice,
				autoscale_weights=autoscale_weights,
				log_warnings=log_warnings,
			)

	def loglike(
			self,
			x=None,
			*,
			start_case=0,
			stop_case=-1,
			step_case=1,
			persist=0,
			leave_out=-1,
			keep_only=-1,
			subsample=-1,
			return_series=True,
			probability_only=False,
	):
		"""
		Compute a log likelihood value and it first derivative.

		Parameters
		----------
		x : {'null', 'init', 'best', array-like, dict, scalar}, optional
			Values for the parameters.  See :ref:`set_values` for details.

		Returns
		-------
		dictx
			The log likelihood is given by key 'll' and the first derivative by key 'dll'.
			Other arrays are also included if `persist` is set to True.

		"""

		if start_case != 0:
			raise NotImplementedError('start_case != 0')
		if stop_case != -1:
			raise NotImplementedError('stop_case != -1')
		if step_case != 1:
			raise NotImplementedError('step_case != 1')
		if leave_out != -1:
			raise NotImplementedError('leave_out != -1')
		if keep_only != -1:
			raise NotImplementedError('keep_only != -1')
		if subsample != -1:
			raise NotImplementedError('subsample != -1')
		if probability_only:
			raise NotImplementedError('probability_only != False')

		from ..util import dictx
		self.__prep_for_compute(x)
		ll2_parts = [m.loglike(persist=persist) for m in self._k_models]
		if not persist:
			result = sum(ll2_parts)
			self._check_if_best(result)
			return result
		ll2 = dictx(
			ll=sum(y.ll for y in ll2_parts),
		)
		if persist & persist_flags.PERSIST_PARTS:
			ll2.ll_parts = [y.ll for y in ll2_parts]
		for key in ll2_parts[0].keys():
			if key != 'll':
				ll2[key] = list(y[key] for y in ll2_parts)
		self._check_if_best(ll2.ll)
		return ll2


	def loglike2(
			self,
			x=None,
			*,
			start_case=0,
			stop_case=-1,
			step_case=1,
			persist=0,
			leave_out=-1,
			keep_only=-1,
			subsample=-1,
			return_series=True,
			probability_only=False,

	):
		"""
		Compute a log likelihood value and it first derivative.

		Parameters
		----------
		x : {'null', 'init', 'best', array-like, dict, scalar}, optional
			Values for the parameters.  See :ref:`set_values` for details.

		Returns
		-------
		dictx
			The log likelihood is given by key 'll' and the first derivative by key 'dll'.
			Other arrays are also included if `persist` is set to True.

		"""

		if start_case != 0:
			raise NotImplementedError('start_case != 0')
		if stop_case != -1:
			raise NotImplementedError('stop_case != -1')
		if step_case != 1:
			raise NotImplementedError('step_case != 1')
		if leave_out != -1:
			raise NotImplementedError('leave_out != -1')
		if keep_only != -1:
			raise NotImplementedError('keep_only != -1')
		if subsample != -1:
			raise NotImplementedError('subsample != -1')
		if probability_only:
			raise NotImplementedError('probability_only != False')

		from ..util import dictx
		self.__prep_for_compute(x)
		ll2_parts = [m.loglike2(persist=persist) for m in self._k_models]
		dll = ll2_parts[0].dll
		for y in ll2_parts[1:]:
			dll = dll.add(y.dll, fill_value=0)
		dll = dll[self.pf.index]
		ll2 = dictx(
			ll=sum(y.ll for y in ll2_parts),
			dll=dll,
		)
		for key in ll2_parts[0].keys():
			if key not in {'ll','dll'}:
				ll2[key] = list(y[key] for y in ll2_parts)
		self._check_if_best(ll2.ll)
		return ll2




from itertools import chain
_RaiseKeyError = object() # singleton for no-default behavior

class SignalDict(dict):  # dicts take a mapping or iterable as their optional first argument
	__slots__ = ('_SignalDict__getitem_callback',
				 '_SignalDict__setitem_callback',
				 '_SignalDict__delitem_callback',) # no __dict__ - that would be redundant
	@staticmethod # because this doesn't make sense as a global function.
	def _process_args(mapping=(), **kwargs):
		if hasattr(mapping, 'items'):
			mapping = getattr(mapping, 'items')()
		return ((k, v) for k, v in chain(mapping, getattr(kwargs, 'items')()))
	def __init__(self, getitem_callback, setitem_callback, delitem_callback, mapping=(), **kwargs):
		self.__getitem_callback = getitem_callback
		self.__setitem_callback = setitem_callback
		self.__delitem_callback = delitem_callback
		super().__init__(self._process_args(mapping, **kwargs))
	def __getitem__(self, k):
		ret = super().__getitem__(k)
		self.__getitem_callback(k)
		return ret
	def __setitem__(self, k, v):
		ret = super().__setitem__(k, v)
		self.__setitem_callback(k,v)
		return ret
	def __delitem__(self, k):
		ret = super().__delitem__(k)
		self.__delitem_callback(k)
		return ret
	def get(self, k, default=None):
		return super().get(k, default)
	def setdefault(self, k, default=None):
		return super().setdefault(k, default)
	def pop(self, k, v=_RaiseKeyError):
		if v is _RaiseKeyError:
			return super().pop(k)
		return super().pop(k, v)
	def update(self, mapping=(), **kwargs):
		super().update(self._process_args(mapping, **kwargs))
	def __contains__(self, k):
		return super().__contains__(k)
	def copy(self): # don't delegate w/ super - dict.copy() -> dict :(
		return type(self)(self)
	def __repr__(self):
		return '{0}({1})'.format(type(self).__name__, super().__repr__())




import yaml
import os
import logging

from addict import Dict as _Dict
from pprint import pformat

__all__ = ('yaml', 'Dict')

def _prettyprint_fallback(x):
	if isinstance(x, _Dict):
		return repr(x)
	else:
		return pformat(x)

class Dict(_Dict):

	def __repr__(self):
		if self.keys():
			m = max(map(len, list(str(_) for _ in self.keys()))) + 1
			return '\n'.join(['┣'+str(k).rjust(m) + ': ' + _prettyprint_fallback(v).replace('\n','\n┃'+' '*(m+2)) for k, v in self.items()])
		else:
			return self.__class__.__name__ + "()"

	@classmethod
	def load(cls, filename, *args, logger=None, encoding='utf-8', Loader=yaml.SafeLoader, **kwargs):
		"""

		Parameters
		----------
		filename : str
			A filename of a yaml file to load, or a yaml file contents as a string.
		args
		allow_json
		kwargs

		Returns
		-------
		Dict
		"""
		from .yaml_checker import yaml_check
		if isinstance(filename, str) and '\n' not in filename:
			if not os.path.exists(filename):
				raise FileNotFoundError(filename)
			try:
				yaml_check(filename, logger=logging.getLogger('') if logger is None else logger)
				with open(filename, 'r', encoding=encoding) as f:
					return cls(yaml.load(f, *(args[1:]), Loader=Loader, **kwargs))
			except Exception as err:
				print("~"*40)
				print(f"ERROR READING {filename}")
				print("~"*40)
				raise

		else:
			return cls(yaml.load(filename, *args, **kwargs))

	@classmethod
	def load_multi(cls, *filenames, **kwargs):
		"""

		Parameters
		----------
		filenames : tuple of str
		kwargs

		Returns
		-------
		Dict
		"""
		filename = filenames[0]
		x = cls.load(filename, **kwargs)
		for filename in filenames[1:]:
			y = cls.load(filename, **kwargs)
			x.update(y)
		return x

	def dump(self, *args, **kwargs):
		from yaml import dump
		if 'default_flow_style' not in kwargs:
			kwargs['default_flow_style'] = False
		if 'indent' not in kwargs:
			kwargs['indent'] = 2
		if len(args) and isinstance(args[0], str):
			if os.path.exists(args[0]):
				raise FileExistsError(args[0])
			dirname = os.path.dirname(args[0])
			if not os.path.exists(dirname):
				os.makedirs(dirname)
			with open(args[0], 'w') as f:
				dump(self.to_dict(), f, *(args[1:]), **kwargs)
		else:
			return dump(self.to_dict(), *args, **kwargs)

	def __xml__(self):
		from xmle import Elem
		x = Elem('div')
		t = x.elem('table', style="margin-top:1px;")
		if len(self):
			tr = t.elem('tr')
			tr.elem('th', text="key")
			tr.elem('th', text='value', style='text-align:left;')
			for k,v in self.items():
				tr = t.elem('tr')
				tr.elem('td', text=str(k))
				try:
					v_ = v.__xml__()
				except AttributeError:
					tr.elem('td', text=str(v), style='text-align:left;')
				else:
					tr.elem('td') << v_
		else:
			tr = t.elem('tr')
			tr.elem('td', text="<empty>")
		return x

	def _repr_html_(self):
		return self.__xml__().tostring()

	def _ipython_display_(self):
		raise AttributeError('_ipython_display_')

	def __getattr__(self, item):
		if item[:5]=="_repr" and item[-1]=="_":
			raise AttributeError(item)
		return self.__getitem__(item)



from pytest import approx, fixture, raises

import larch
import pandas
from larch import data_warehouse
from larch.roles import P,X

@fixture(scope="module")
def swissmetro_raw_df():
	raw_df = pandas.read_csv(data_warehouse.example_file('swissmetro.csv.gz'))
	raw_df['SM_COST'] = raw_df['SM_CO'] * (raw_df["GA"]==0)
	raw_df['TRAIN_COST'] = raw_df.eval("TRAIN_CO * (GA == 0)")
	raw_df['CAR_AV_SP'] = raw_df.eval("CAR_AV * (SP!=0)")
	raw_df['TRAIN_AV_SP'] = raw_df.eval("TRAIN_AV * (SP!=0)")
	keep = raw_df.eval("PURPOSE in (1,3) and CHOICE != 0")
	return raw_df[keep]


def test_latent_class(swissmetro_raw_df):

	dfs = larch.DataFrames(swissmetro_raw_df, alt_codes=[1,2,3])

	m1 = larch.Model(dataservice=dfs)
	m1.availability_co_vars = {
		1: "TRAIN_AV_SP",
		2: "SM_AV",
		3: "CAR_AV_SP",
	}
	m1.choice_co_code = 'CHOICE'
	m1.utility_co[1] = P("ASC_TRAIN") + X("TRAIN_CO*(GA==0)") * P("B_COST")
	m1.utility_co[2] = X("SM_CO*(GA==0)") * P("B_COST")
	m1.utility_co[3] = P("ASC_CAR") + X("CAR_CO") * P("B_COST")


	m2 = larch.Model(dataservice=dfs)
	m2.availability_co_vars = {
		1: "TRAIN_AV_SP",
		2: "SM_AV",
		3: "CAR_AV_SP",
	}
	m2.choice_co_code = 'CHOICE'
	m2.utility_co[1] = P("ASC_TRAIN") + X("TRAIN_TT") * P("B_TIME") + X("TRAIN_CO*(GA==0)") * P("B_COST")
	m2.utility_co[2] = X("SM_TT") * P("B_TIME") + X("SM_CO*(GA==0)") * P("B_COST")
	m2.utility_co[3] = P("ASC_CAR") + X("CAR_TT") * P("B_TIME") + X("CAR_CO") * P("B_COST")

	km = larch.Model()
	km.utility_co[2] = P.W_OTHER

	from larch.model.latentclass import LatentClassModel
	m = LatentClassModel(km, {1:m1, 2:m2})

	m.load_data()

	m.set_value(P.ASC_CAR, 0.125/2)
	m.set_value(P.ASC_TRAIN, -0.398/2)
	m.set_value(P.B_COST, -.0126/2)
	m.set_value(P.B_TIME, -0.028/2)
	m.set_value(P.W_OTHER, 1.095/2)

	check1 = m.check_d_loglike()

	assert dict(check1.data.analytic) == approx({
		'ASC_CAR': -81.69736186616234,
		'ASC_TRAIN': -613.131371089499,
		'B_COST': -6697.31706964777,
		'B_TIME': -40104.940072046316,
		'W_OTHER': 245.43145056623683,
	})

	assert check1.data.similarity.min() > 4

	m.set_value(P.ASC_CAR, 0.125)
	m.set_value(P.ASC_TRAIN, -0.398)
	m.set_value(P.B_COST, -.0126)
	m.set_value(P.B_TIME, -0.028)
	m.set_value(P.W_OTHER, 1.095)

	assert m.loglike() == approx(-5208.502259337974)

	check2 = m.check_d_loglike()

	assert dict(check2.data.analytic) == approx({
		'ASC_CAR': 0.6243716033364302,
		'ASC_TRAIN': 0.9297965389102578,
		'B_COST': -154.03997923797007,
		'B_TIME': 76.19297915128493,
		'W_OTHER': -0.7936963902343083,
	})

	assert check2.data.similarity.min() > 2 # similarity is a bit lower very close to the optimum


def test_latent_class_mixed_data(swissmetro_raw_df):

	dfs = larch.DataFrames(swissmetro_raw_df, alt_codes=[1,2,3])

	m1 = larch.Model(dataservice=dfs)
	m1.availability_co_vars = {
		1: "TRAIN_AV_SP",
		2: "SM_AV",
		3: "CAR_AV_SP",
	}
	m1.choice_co_code = 'CHOICE'
	m1.utility_co[1] = P("ASC_TRAIN") + X("TRAIN_CO*(GA==0)") * P("B_COST")
	m1.utility_co[2] = X("SM_CO*(GA==0)") * P("B_COST")
	m1.utility_co[3] = P("ASC_CAR") + X("CAR_CO") * P("B_COST")


	m2 = larch.Model(dataservice=dfs)
	m2.availability_co_vars = {
		1: "TRAIN_AV_SP",
		2: "SM_AV",
		3: "CAR_AV_SP",
	}
	m2.choice_co_code = 'CHOICE'
	m2.utility_co[1] = P("ASC_TRAIN") + X("TRAIN_TT") * P("B_TIME") + X("TRAIN_CO*(GA==0)") * P("B_COST")
	m2.utility_co[2] = X("SM_TT") * P("B_TIME") + X("SM_CO*(GA==0)") * P("B_COST")
	m2.utility_co[3] = P("ASC_CAR") + X("CAR_TT") * P("B_TIME") + X("CAR_CO") * P("B_COST")

	dfs2 = larch.DataFrames(swissmetro_raw_df, alt_codes=[1,2])
	km = larch.Model(dataservice=dfs2)
	km.utility_co[2] = P.W_OTHER

	from larch.model.latentclass import LatentClassModel
	with raises(ValueError):
		m = LatentClassModel(km, {1:m1, 2:m2})

def test_latent_class_full_data(swissmetro_raw_df):

	dfs = larch.DataFrames(swissmetro_raw_df, alt_codes=[1,2,3])

	m1 = larch.Model(dataservice=dfs)
	m1.availability_co_vars = {
		1: "TRAIN_AV_SP",
		2: "SM_AV",
		3: "CAR_AV_SP",
	}
	m1.choice_co_code = 'CHOICE'
	m1.utility_co[1] = P("ASC_TRAIN") + X("TRAIN_CO*(GA==0)") * P("B_COST")
	m1.utility_co[2] = X("SM_CO*(GA==0)") * P("B_COST")
	m1.utility_co[3] = P("ASC_CAR") + X("CAR_CO") * P("B_COST")


	m2 = larch.Model(dataservice=dfs)
	m2.availability_co_vars = {
		1: "TRAIN_AV_SP",
		2: "SM_AV",
		3: "CAR_AV_SP",
	}
	m2.choice_co_code = 'CHOICE'
	m2.utility_co[1] = P("ASC_TRAIN") + X("TRAIN_TT") * P("B_TIME") + X("TRAIN_CO*(GA==0)") * P("B_COST")
	m2.utility_co[2] = X("SM_TT") * P("B_TIME") + X("SM_CO*(GA==0)") * P("B_COST")
	m2.utility_co[3] = P("ASC_CAR") + X("CAR_TT") * P("B_TIME") + X("CAR_CO") * P("B_COST")

	km = larch.Model(dataservice=dfs, alts=[1,2])
	km.utility_co[2] = P.W_OTHER

	from larch.model.latentclass import LatentClassModel
	m = LatentClassModel(km, {1:m1, 2:m2})

	m.load_data()

	m.set_value(P.ASC_CAR, 0.125/2)
	m.set_value(P.ASC_TRAIN, -0.398/2)
	m.set_value(P.B_COST, -.0126/2)
	m.set_value(P.B_TIME, -0.028/2)
	m.set_value(P.W_OTHER, 1.095/2)

	check1 = m.check_d_loglike()

	assert dict(check1.data.analytic) == approx({
		'ASC_CAR': -81.69736186616234,
		'ASC_TRAIN': -613.131371089499,
		'B_COST': -6697.31706964777,
		'B_TIME': -40104.940072046316,
		'W_OTHER': 245.43145056623683,
	})



import os
from pathlib import Path
import tables as tb
import numpy
import pandas
import warnings
from ....util import Dict
from ....util.aster import asterize
from ....util.text_manip import truncate_path_for_display


from ... import _reserved_names_

from .generic import H5Pod, NoKnownShape, IncompatibleShape
from .idca import H5PodCA
from .idco import H5PodCO
from .idcs import H5PodCS
from .idce import H5PodCE
from .idga import H5PodGA
from .idrc import H5PodRC
from .id0a import H5Pod0A

def H5PodFactory(uri:str):
	from urllib.parse import urlparse, parse_qs
	p = urlparse(uri, scheme='file')
	q = parse_qs(p.query)
	if 'type' in q:
		cls = _pod_types[q['type'][0]]
	else:
		cls = H5Pod
	if 'mode' in q:
		mode = q['mode'][0]
	else:
		mode = 'r'
	return cls(filename=p.path, mode=mode, groupnode=p.fragment)


_pod_types = {
	'idco': H5PodCO,
	'idca': H5PodCA,
	'idce': H5PodCE,
	'idga': H5PodGA,
	'idcs': H5PodCS,
	'co': H5PodCO,
	'ca': H5PodCA,
	'ce': H5PodCE,
	'ga': H5PodGA,
	'cs': H5PodCS,
}




class TouchNotify():
	"""
	A mixin to execute a notification callback any time a
	mutate-possible command is called.
	"""

	def __init__(self, *args, touch_callback=None, **kwargs):
		self.__touch_callback = touch_callback if callable(touch_callback) else lambda: None
		super().__init__(*args, **kwargs)

	def get_touch_callback(self):
		try:
			return self.__touch_callback
		except AttributeError:
			return lambda: None

	def set_touch_callback(self, callback):
		if callable(callback):
			self.__touch_callback = callback
		elif callback is None:
			self.__touch_callback = lambda: None
		else:
			raise TypeError('callback must be callable')

	def del_touch_callback(self):
		self.__touch_callback = lambda: None

	def touch(self):
		self.__touch_callback()

	def __setitem__(self, key, value):
		ret = super().__setitem__(key, value)
		try:
			self.__touch_callback()
		except AttributeError:
			pass
		return ret

	def __setattr__(self, key, value):
		ret = super().__setattr__(key, value)
		try:
			self.__touch_callback()
		except AttributeError:
			pass
		return ret

	def __delitem__(self, key):
		ret = super().__delitem__(key)
		self.__touch_callback()
		return ret

	def __delattr__(self, key):
		ret = super().__delattr__(key)
		self.__touch_callback()
		return ret

	def setdefault(self, *args, **kwargs):
		ret = super().setdefault(*args, **kwargs)
		self.__touch_callback()
		return ret

	def append(self, *args, **kwargs):
		ret = super().append(*args, **kwargs)
		self.__touch_callback()
		return ret

	def extend(self, *args, **kwargs):
		ret = super().extend(*args, **kwargs)
		self.__touch_callback()
		return ret

	def update(self, *args, **kwargs):
		ret = super().update(*args, **kwargs)
		self.__touch_callback()
		return ret

	def insert(self, *args, **kwargs):
		ret = super().insert(*args, **kwargs)
		self.__touch_callback()
		return ret

	def remove(self, *args, **kwargs):
		ret = super().remove(*args, **kwargs)
		self.__touch_callback()
		return ret

	def pop(self, *args, **kwargs):
		ret = super().pop(*args, **kwargs)
		self.__touch_callback()
		return ret

	def popitem(self, *args, **kwargs):
		ret = super().popitem(*args, **kwargs)
		self.__touch_callback()
		return ret

	def clear(self, *args, **kwargs):
		ret = super().clear(*args, **kwargs)
		self.__touch_callback()
		return ret

	def __call__(self, *args, **kwargs):
		ret = super().__call__(*args, **kwargs)
		self.__touch_callback()
		return ret

	def __iadd__(self, other):
		ret = super().__iadd__(other)
		self.__touch_callback()
		return ret

	def __isub__(self, other):
		ret = super().__isub__(other)
		self.__touch_callback()
		return ret

	def __imul__(self, other):
		ret = super().__imul__(other)
		self.__touch_callback()
		return ret

	def __idiv__(self, other):
		ret = super().__idiv__(other)
		self.__touch_callback()
		return ret

	def __iand__(self, other):
		ret = super().__iand__(other)
		self.__touch_callback()
		return ret

	def __ior__(self, other):
		ret = super().__ior__(other)
		self.__touch_callback()
		return ret

	def __ixor__(self, other):
		ret = super().__ixor__(other)
		self.__touch_callback()
		return ret

	def __getstate__(self):
		return {}

	def __setstate__(self, state):
		self.__touch_callback = lambda: None

#
# class EditableList:
#
# 	def __init__(self, content=None, touched=None):
# 		self._data = [] if content is None else list(content)
# 		if touched is None:
# 			self._touched = lambda: print("TOUCH")
# 		else:
# 			self._touched = touch
#
# 	def __getitem__(self, position):
# 		return self._data[position]
#
# 	def __setitem__(self, position, value):
# 		self._data[position] = value
# 		self._touched()
#
# 	def append(self, value):
# 		self._data.append(value)
# 		self._touched()
#
#
# # this is our descriptor object
# class Bar(object):
# 	def __init__(self):
# 		self.values = EditableList()
#
# 	def __get__(self, instance, owner):
# 		print("returned from descriptor object", type(instance), type(owner), instance.its_me())
# 		return self.values
#
# 	def __set__(self, instance, values):
# 		print("set in descriptor object", type(instance), type(values), instance.its_me())
# 		self.values = EditableList(values)
# 		self.values._touched = instance.mangle
#
# 	def __delete__(self, instance):
# 		print("deleted in descriptor object", type(instance))
# 		del self.values
# 		self.values = EditableList()
#
# 	def __set_name__(self, owner, name):
# 		print("setname in descriptor object", type(owner), owner, name, type(self))
# 		self.name = name
# 		self.internal_name = '_' + name
#
#
# class Foo(object):
# 	bar = Bar()
# 	bax = 888.888
#
# 	def __init__(self):
# 		self.bar = []
#
# 	def mangle(self):
# 		print("MANGLED")
#
# 	def its_me(self):
# 		return repr(self.bax)


import numpy
from .idco import H5PodCO
from ...general import selector_len_for, _sqz_same

class H5PodRC(H5PodCO):

	def __init__(self, rowindexes, colindexes, *args, **kwargs):
		super().__init__(*args, **kwargs)
		if isinstance(rowindexes, str):
			self.rowindexes = self._groupnode._v_children[rowindexes][:]
		elif isinstance(rowindexes, numpy.ndarray):
			self.rowindexes = rowindexes
		elif isinstance(rowindexes, int) and isinstance(colindexes, (str, numpy.ndarray)):
			self.rowindexes = None  # fix in a moment
		else:
			raise TypeError('rowindexes must be str, int, or ndarray')

		if isinstance(colindexes, str):
			self.colindexes = self._groupnode._v_children[colindexes][:]
		elif isinstance(colindexes, numpy.ndarray):
			self.colindexes = colindexes
		elif isinstance(colindexes, int) and self.rowindexes is not None:
			self.colindexes = numpy.full_like(self.rowindexes, colindexes, dtype=int)
		else:
			raise TypeError('colindexes must be str, int, or ndarray')

		if self.rowindexes is None:
			self.rowindexes = numpy.full_like(self.colindexes, rowindexes, dtype=int)

	@property
	def podtype(self):
		return 'idrc'


	def __getitem__(self, item):

		if isinstance(item, tuple) and len(item)>=2 and isinstance(item[-1], slice):
			names, slice_ = item[:-1], item[-1]
		else:
			names = item
			slice_ = None

		# convert a single name string to a one item list
		if isinstance(names, str):
			names = [names, ]

		dtype = numpy.float64

		result = numpy.zeros( [selector_len_for(slice_, self.shape[0]), *self.shape[1:], len(names)], dtype=dtype)

		for i, cmd in enumerate(names):
			temp = self._evaluate_single_item(cmd, None)
			if slice_ is not None:
				result[...,i] = temp[self.rowindexes[slice_],self.colindexes[slice_]]
			else:
				result[...,i] = temp[self.rowindexes,self.colindexes]
		return result

	def _load_into(self, names, slc, result):
		if len(self._altcodes) == 0:
			raise ValueError("alternatives vector not set")
		# convert a single name string to a one item list
		if isinstance(names, str):
			names = [names, ]
		names = [self._get_list_from_bunchname(i) for i in names]

		# check that all names are actually lists or tuples of the same length
		first_len = len(names[0])
		for cmd in names[1:]:
			assert( len(cmd) == first_len )

		assert (tuple(result.shape) == tuple([selector_len_for(slc, self.shape[0]), *self.shape[1:], len(names)]))
		for i, cmd in enumerate(names):
			temp = self._evaluate_single_item(cmd, None)
			if slc is not None:
				result[...,i] = temp[self.rowindexes[slc],self.colindexes[slc]]
			else:
				result[...,i] = temp[self.rowindexes,self.colindexes]
		return result

	@property
	def shape(self):
		"""The shape of the pod.

		"""
		return tuple(self.rowindexes.shape)

	@property
	def metashape(self):
		"""The shape of the underlying skims.

		"""
		return tuple(self._groupnode._v_attrs.SHAPE)

	def load_data_item(self, name, result, selector=None):
		"""Load a slice of the pod arrays into an array in memory"""
		# convert a single name string to a one item list

		from ...general import _sqz_same
		_sqz_same(result.shape, [selector_len_for(selector, self.shape[0]), *self.shape[1:]])

		temp = self._evaluate_single_item(name, None)
		if selector is not None:
			result[:] = temp[self.rowindexes[selector], self.colindexes[selector]]
		else:
			result[:] = temp[self.rowindexes, self.colindexes]
		return result

	def as_idca(self):
		ret = H5PodRCasCA(filename=self, rowindexes=self.rowindexes, colindexes=self.colindexes, ident=self.ident+"_as_idca")
		return ret

class H5PodRCasCA(H5PodRC):

	@property
	def podtype(self):
		return 'idca'

	def load_data_item(self, name, result, selector=None):
		"""Load a slice of the pod arrays into an array in memory"""
		# convert a single name string to a one item list

		from ...general import _sqz_same
		_sqz_same(result.shape, [selector_len_for(selector, self.shape[0]), *self.shape[1:]])

		temp = self._evaluate_single_item(name, None)
		if selector is not None:
			result[:,:] = temp[self.rowindexes[selector], self.colindexes[selector]][:,None]
		else:
			result[:,:] = temp[self.rowindexes, self.colindexes][:,None]
		return result

	@property
	def shape(self):
		"""The shape of the pod.

		"""
		return super().shape + (self.trailing_dim,)

	@property
	def trailing_dim(self):
		try:
			return self._trailing_dim
		except AttributeError:
			return -1

	@trailing_dim.setter
	def trailing_dim(self, value):
		self._trailing_dim = int(value)



#
# from ..data_services.examples import example_file
# from ..data_services import H5PodCE, DataService
# from ..model import Model
# import numpy
# from pytest import approx
# import pandas
# from ..roles import P, X, PX
# from collections import OrderedDict
#
#
# def test_idce_mnl():
# 	f = example_file('MTCwork.csv.gz')
#
# 	ce = H5PodCE.from_csv(f)
#
# 	ce.create_indexes_from_labels('case_ix', 'casenum')
# 	ce.create_indexes_from_labels('alt_ix', 'altnum')
#
# 	ce.set_casealt_indexes('case_ix', 'alt_ix')
#
# 	dx = DataService(
# 		ce,
# 		altids=[1, 2, 3, 4, 5, 6],
# 		altnames=['DA', 'S2', 'S3', 'TR', 'BK', 'WK'],
# 	)
#
# 	av = ce.as_array_idca('1', present='1', dtype=numpy.bool)
#
# 	assert( av.shape == (5029, 6, 1) )
#
# 	ref = Model.Example()
#
# 	ref.set_values(
# 		totcost=-0.01,
# 		tottime=-0.01,
# 		ASC_BIKE=-1,
# 		ASC_SR2=-1,
# 	)
#
# 	ref.load_data()
#
# 	ref_ll = ref.loglike()
#
# 	assert( approx(-8739.682333540684) == ref_ll )
#
# 	ref_dll = ref.d_loglike()
#
# 	dx.add_pod(ref.dataservice.idco[0], broadcastable=False)
#
# 	m = Model(dataservice=dx)
#
# 	m.utility_co[2] = P("ASC_SR2") + P("hhinc#2") * X("hhinc")
# 	m.utility_co[3] = P("ASC_SR3P") + P("hhinc#3") * X("hhinc")
# 	m.utility_co[4] = P("ASC_TRAN") + P("hhinc#4") * X("hhinc")
# 	m.utility_co[5] = P("ASC_BIKE") + P("hhinc#5") * X("hhinc")
# 	m.utility_co[6] = P("ASC_WALK") + P("hhinc#6") * X("hhinc")
#
# 	m.utility_ca = PX("tottime") + PX("totcost")
#
# 	m.availability_var = '_avail_'
# 	m.choice_ca_var = 'chose'
#
# 	m.title = "MTC Example 1 (Simple MNL)"
#
# 	m.set_values(
# 		totcost=-0.01,
# 		tottime=-0.01,
# 		ASC_BIKE=-1,
# 		ASC_SR2=-1,
# 	)
#
# 	m.load_data()
#
# 	m_ll = m.loglike(cache_clear=True)
#
# 	assert( approx(-8739.682333540684) == m_ll )
#
# 	m_dll = m.d_loglike(return_series=True)
#
# 	q = pandas.DataFrame.from_dict(OrderedDict([
# 		('ref', ref_dll),
# 		('ce', m_dll),
# 	]))
# 	q['diff'] = q.ref - q.ce
#
# 	for row in q.itertuples():
# 		assert( approx(row.ref) == row.ce )
#
#
# def test_idce_qmnl():
# 	f = example_file('MTCwork.csv.gz')
#
# 	ce = H5PodCE.from_csv(f)
#
# 	ce.create_indexes_from_labels('case_ix', 'casenum')
# 	ce.create_indexes_from_labels('alt_ix', 'altnum')
#
# 	ce.set_casealt_indexes('case_ix', 'alt_ix')
#
# 	dx = DataService(
# 		ce,
# 		altids=[1, 2, 3, 4, 5, 6],
# 		altnames=['DA', 'S2', 'S3', 'TR', 'BK', 'WK'],
# 	)
#
# 	av = ce.as_array_idca('1', present='1', dtype=numpy.bool)
#
# 	assert( av.shape == (5029, 6, 1) )
#
# 	ref = Model.Example()
# 	ref.quantity_ca = P.Q1 + P.Qaltnum * X.altnum
# 	ref.set_values(
# 		totcost=-0.01,
# 		tottime=-0.01,
# 		ASC_BIKE=-1,
# 		ASC_SR2=-1,
# 		Q1=-1,
# 		Qaltnum=0.1,
# 	)
#
# 	ref.load_data()
#
# 	ref_ll = ref.loglike()
#
# 	assert( approx(-11877.84511534832) == ref_ll )
#
# 	ref_dll = ref.d_loglike()
#
# 	dx.add_pod(ref.dataservice.idco[0], broadcastable=False)
#
# 	m = Model(dataservice=dx)
#
# 	m.utility_co[2] = P("ASC_SR2") + P("hhinc#2") * X("hhinc")
# 	m.utility_co[3] = P("ASC_SR3P") + P("hhinc#3") * X("hhinc")
# 	m.utility_co[4] = P("ASC_TRAN") + P("hhinc#4") * X("hhinc")
# 	m.utility_co[5] = P("ASC_BIKE") + P("hhinc#5") * X("hhinc")
# 	m.utility_co[6] = P("ASC_WALK") + P("hhinc#6") * X("hhinc")
#
# 	m.utility_ca = PX("tottime") + PX("totcost")
# 	m.quantity_ca = P.Q1 + P.Qaltnum * X.altnum
# 	m.availability_var = '_avail_'
# 	m.choice_ca_var = 'chose'
#
# 	m.title = "MTC Example 1 (Simple MNL)"
#
# 	m.set_values(
# 		totcost=-0.01,
# 		tottime=-0.01,
# 		ASC_BIKE=-1,
# 		ASC_SR2=-1,
# 		Q1=-1,
# 		Qaltnum=0.1,
# 	)
#
# 	m.load_data()
#
# 	m_ll = m.loglike(cache_clear=True)
#
# 	assert( approx(-11877.84511534832) == m_ll )
#
# 	m_dll = m.d_loglike(return_series=True)
#
# 	q = pandas.DataFrame.from_dict(OrderedDict([
# 		('ref', ref_dll),
# 		('ce', m_dll),
# 	]))
# 	q['diff'] = q.ref - q.ce
#
# 	correct = {
# 		'hhinc#6':     -24910.315308,
# 		'hhinc#2':       3398.416194,
# 		'hhinc#5':     -19911.378054,
# 		'hhinc#3':    -134912.951431,
# 		'ASC_WALK':      -431.000440,
# 		'ASC_BIKE':      -320.626546,
# 		'ASC_TRAN':      -293.301888,
# 		'hhinc#4':     -19475.950414,
# 		'totcost':     344656.334682,
# 		'ASC_SR3P':     -2275.752299,
# 		'ASC_SR2':         75.169003,
# 		'tottime':     -52936.139683,
# 		'Q1':             528.768035,
# 		'Qaltnum':       -528.768035,
# 	}
#
# 	for row in q.itertuples():
# 		assert( approx(row.ref) == row.ce )
# 		assert( approx(row.ref) == correct[row.Index] )
#
#
#





from selenium import webdriver
import os
from PIL import Image, ImageDraw, ImageEnhance, ImageChops

import base64
from io import BytesIO



class _ScreenCapper:

	def __init__(self):
		options = webdriver.ChromeOptions()
		options.add_argument('headless')
		options.add_argument('window-size=1024x768')
		options.add_argument("disable-overlay-scrollbar")
		self.driver = webdriver.Chrome(chrome_options=options)

	def __del__(self):
		self.driver.close()

	def _screenshot_formatter(self, url, window_size=(1024, 768), full_height=False):
		self.driver.get(url)
		if window_size[1] is None or full_height:
			# Full height window
			self.driver.maximize_window()
			height = self.driver.execute_script("return document.body.scrollHeight")
			window_size = window_size[0], height
			self.driver.set_window_size(*window_size)

	def screenshot_to_file(self, url, file_name, log=None, output_dir='./', window_size=(1024, 768), full_height=False):
		if log is not None:
			log("Capturing", url, "screenshot as", file_name, "...")
		if not os.path.exists(output_dir):
			os.makedirs(output_dir)
		self._screenshot_formatter(url, window_size, full_height=full_height)
		self.driver.save_screenshot(file_name)
		if log is not None:
			log("Done capturing", url)

	def screenshot_to_base64(self, url, log=None, window_size=(1024, 768), full_height=False):
		if log is not None:
			log("Capturing", url, "screenshot as base64...")
		self._screenshot_formatter(url, window_size, full_height=full_height)
		result = self.driver.get_screenshot_as_base64()
		if log is not None:
			log("Done capturing", url)
		return result

_grabber = None


def screenshot(url, file_name=None, log=None, output_dir='./', window_size=(1024, 768), full_height=False):
	global _grabber
	if _grabber is None:
		_grabber = _ScreenCapper()
	if file_name is None:
		return _grabber.screenshot_to_base64(url, log=log, window_size=window_size, full_height=full_height)
	else:
		_grabber.screenshot_to_file(url, file_name, log=log, output_dir=output_dir, window_size=window_size, full_height=full_height)

# https://gist.github.com/rinchik/d023578a705d6d4a5b12e235c5a9df9a#file-averageregionbrightness-py

def _process_region(image, x, y, width, height, tolerance=100):
	region_total = 0

	# This can be used as the sensitivity factor, the larger it is the less sensitive the comparison
	factor = tolerance

	for coordinateY in range(y, y + height):
		for coordinateX in range(x, x + width):
			try:
				pixel = image.getpixel((coordinateX, coordinateY))
				region_total += sum(pixel) / 4
			except:
				return

	return region_total / factor


def diff_images(imagefile1, imagefile2, return_blocked1=False, tolerance=100):
	'''

	Parameters
	----------
	imagefile1, imagefile2 : images
		Two images to compare.  They should be the same size.  Give by filename or with
		raw data.
	return_blocked1 : bool, default False
		If true, also return an image `Elem` based on imagefile1 showing where the mismatch blocks are.
	tolerance : int
		This can be used as the sensitivity factor, the larger it is the less sensitive the comparison

	Returns
	-------
	int
		The number of mismatched blocks
	Elem
		optional, see return_blocked1
	'''
	screenshot_staging = Image.open(imagefile1)
	screenshot_production = Image.open(imagefile2)
	columns = 60
	rows = 80
	screen_width, screen_height = screenshot_staging.size

	block_width = ((screen_width - 1) // columns) + 1  # this is just a division ceiling
	block_height = ((screen_height - 1) // rows) + 1

	mismatch_blocks = 0

	if return_blocked1:
		degenerate = Image.new(screenshot_staging.mode, screenshot_staging.size, '#FFFFFF')

		if 'A' in screenshot_staging.getbands():
			degenerate.putalpha(screenshot_staging.getchannel('A'))

		screenshot_staging_1 = Image.blend(degenerate, screenshot_staging, 0.25)



	for y in range(0, screen_height, block_height + 1):
		for x in range(0, screen_width, block_width + 1):
			region_staging = _process_region(screenshot_staging, x, y, block_width, block_height, tolerance=tolerance)
			region_production = _process_region(screenshot_production, x, y, block_width, block_height, tolerance=tolerance)

			if region_staging is not None and region_production is not None and region_production != region_staging:
				mismatch_blocks += 1
				if return_blocked1:
					draw = ImageDraw.Draw(screenshot_staging_1)
					draw.rectangle((x, y, x + block_width, y + block_height), outline="red")

	if return_blocked1:
		buffered = BytesIO()
		screenshot_staging_1.save(buffered, format="PNG")
		from xmle import Elem
		blocked1 = Elem.from_png_raw(buffered.getvalue())
		return mismatch_blocks, blocked1

	return mismatch_blocks




# https://stackoverflow.com/questions/10615901/trim-whitespace-using-pil/10616717

def trim_uniform_border(im):
	"""
	Trim a border from

	Parameters
	----------
	im : PIL.Image.Image

	Returns
	-------
	PIL.Image.Image

	"""
	bg = Image.new(im.mode, im.size, im.getpixel((0,0)))
	diff = ImageChops.difference(im, bg)
	diff = ImageChops.add(diff, diff, 2.0, -100)
	bbox = diff.getbbox()
	if bbox:
		return im.crop(bbox)





# Errors

class MissingDataError(ValueError):
	pass

class DuplicateColumnNames(ValueError):
	pass

class BHHHSimpleStepFailure(RuntimeError):
	pass

# Warnings

class ParameterNotInModelWarning(UserWarning):
	pass




#!/usr/bin/env python
import os
import zipfile
import hashlib
import time


def _rec_split(s):
	rest, tail = os.path.split(s)
	if rest in ('', os.path.sep):
		return tail,
	return _rec_split(rest) + (tail,)


def _any_dot(s):
	for i in _rec_split(s):
		if len(i) > 0 and i[0] == '.':
			return True
	return False


def _zipdir(path, ziph, skip_dots=True, extra_layer=True, log=print):
	# ziph is zipfile handle
	keep_dots = not skip_dots
	for root, dirs, files in os.walk(path):
		folder = os.path.basename(root)
		if keep_dots or not _any_dot(folder):
			log(f'zipping folder: {folder} in {root}')
			for file in files:
				if keep_dots or not _any_dot(file):
					ziph.write(os.path.join(root, file), os.path.relpath(os.path.join(root, file), os.path.join(path,
																												'..' if extra_layer else '.')))
		else:
			log(f'not zipping folder: {folder} in {root}')


def zipdir(source_dir, zip_file_name=None, skip_dots=True, extra_layer=False, log=print):
	"""

	Parameters
	----------
	source_dir
	zip_file_name : str
		If not given, uses the name of the sourcedir.
	skip_dots : bool, defaults True
		Ignore files and dirs that start with a dot.

	Returns
	-------
	str
		zip_file_name
	"""
	if zip_file_name is None:
		if source_dir[-1] in ('/', '\\'):
			usepath = source_dir[:-1]
		else:
			usepath = source_dir
		zip_file_name = usepath + '.zip'
	with zipfile.ZipFile(zip_file_name, 'w', zipfile.ZIP_DEFLATED) as zipf:
		_zipdir(source_dir, zipf, skip_dots=skip_dots, extra_layer=extra_layer, log=log)
	return zip_file_name


def zipmod(module, zip_file_name, skip_dots=True):
	"""
	Create a zipfile from a module

	Parameters
	----------
	module
	zip_file_name
	skip_dots

	Returns
	-------

	"""
	with zipfile.ZipFile(zip_file_name, 'w', zipfile.ZIP_DEFLATED) as zipf:
		_zipdir(module.__path__[0], zipf, skip_dots=skip_dots)


def zipmod_temp(module, skip_dots=True):
	import tempfile
	tempdir = tempfile.TemporaryDirectory()
	zip_file_name = os.path.join(tempdir.name, module.__name__ + ".zip")
	zipmod(module, zip_file_name, skip_dots=skip_dots)
	return zip_file_name, tempdir


def make_hash_file(fname):
	hash256 = hashlib.sha256()
	if fname[-3:] == '.gz':
		import gzip
		with gzip.open(fname, "rb") as f:
			for chunk in iter(lambda: f.read(4096), b""):
				hash256.update(chunk)
		h = hash256.hexdigest()
		with open(fname[:-3] + ".sha256.txt", "w") as fh:
			fh.write(h)
	else:
		with open(fname, "rb") as f:
			for chunk in iter(lambda: f.read(4096), b""):
				hash256.update(chunk)
		h = hash256.hexdigest()
		with open(fname + ".sha256.txt", "w") as fh:
			fh.write(h)


def verify_hash_file(fname, hash_dir=None, max_retries=5):
	hash256 = hashlib.sha256()

	retries = 0
	while retries < max_retries:
		try:
			with open(fname, "rb") as f:
				for chunk in iter(lambda: f.read(4096), b""):
					hash256.update(chunk)
		except PermissionError:
			import time
			time.sleep(5)
			retries += 1
		except:
			raise
		else:
			break

	h = hash256.hexdigest()
	if hash_dir is None:
		with open(fname + ".sha256.txt", "r") as fh:
			h_x = fh.read()
	else:
		with open(os.path.join(hash_dir, os.path.basename(fname) + ".sha256.txt"), "r") as fh:
			h_x = fh.read()
	if h != h_x:
		if hash_dir:
			raise ValueError(f"bad hash on {fname} with hash_dir={hash_dir}")
		else:
			raise ValueError(f"bad hash on {fname}")


def archive_subdir(destination_dir):
	import time, os
	destination_dir = os.path.join(destination_dir, time.strftime("%Y_%m_%d-%H_%M"))
	os.makedirs(destination_dir, exist_ok=True)
	return destination_dir

def gzip_dir(source_dir, patterns=("*.*",), make_hash=True, exclude=".sha256.txt", destination_dir=None, archive_subdir=False, remove_original=True, log=None):
	"""Individually gzip every file matching patterns in source_dir."""

	import gzip, glob
	import shutil, os, time

	if destination_dir is None:
		destination_dir = source_dir
	else:
		os.makedirs(destination_dir, exist_ok=True)

	if archive_subdir:
		if isinstance(archive_subdir, str):
			destination_dir = os.path.join(destination_dir, archive_subdir)
		else:
			destination_dir = os.path.join(destination_dir, time.strftime("%Y_%m_%d-%H_%M"))
		os.makedirs(destination_dir, exist_ok=True)

	targets = set()

	for pattern in patterns:
		for f in glob.glob(os.path.join(source_dir, pattern)):
			if exclude in f:
				continue  # don't re-gzip the hash files by default
			targets.add(f)

	for f in targets:
		if make_hash:
			make_hash_file(f)
		f_relative = os.path.relpath(f, start=source_dir)

		if f[-3:] != '.gz':
			with open(f, 'rb') as f_in:
				f_to = os.path.join(destination_dir, f_relative)
				with gzip.open( f_to + '.gz', 'wb') as f_out:
					if log is not None:
						log(f"{f} --> {f_to}.gz")
					shutil.copyfileobj(f_in, f_out)
			if remove_original:
				os.remove(f)

def gzip_archive(source_dir, patterns=("*.*",), make_hash=False, exclude=".sha256.txt", destination_dir=None, archive_subdir=True, remove_original=False, log=None):
	"""Same as gzip_dir but different default inputs.

	Parameters
	----------
	source_dir
	patterns
	make_hash
	exclude
	destination_dir
	archive_subdir
	remove_original

	Returns
	-------

	"""
	return gzip_dir(source_dir=source_dir, patterns=patterns,
					make_hash=make_hash, exclude=exclude,
					destination_dir=destination_dir, archive_subdir=archive_subdir,
					remove_original=remove_original,
					log=log)



class Archiver():

	def __init__(self, *outdirs, log=None, handle=None):
		if handle is None:
			self._handle = time.strftime("%Y_%m_%d-%H_%M")
		else:
			self._handle = handle
		self.outdirs = []
		self.archive_dirs = []
		for od in outdirs:
			self.add_outdir(od)
		if log is None:
			from .logs.normal import log as _log
			self.log = _log.info
		elif log is False:
			self.log = None
		else:
			self.log = log

	def add_outdir(self, outdir):
		self.outdirs.append(outdir)
		d = os.path.join(outdir, self._handle)
		os.makedirs(d, exist_ok=True)
		self.archive_dirs.append(d)

	def gzip_by_glob(self, pattern, source_dir='.'):
		for arch in self.archive_dirs:
			gzip_archive(
				source_dir=source_dir,
				patterns=(pattern,),
				destination_dir=arch,
				archive_subdir=False,
				log=self.log,
			)

	def zip_dir(self, source_dir, zip_file_name=None, **kwargs):
		if zip_file_name is None:
			zip_file_name = os.path.basename(source_dir)+".zip"
		for arch in self.archive_dirs:
			zipdir(
				source_dir=source_dir,
				zip_file_name=os.path.join(arch, zip_file_name),
				log=self.log,
				**kwargs
			)

	zip_by_dir = zip_dir


import pandas, numpy
from xmle import Elem
from joblib import Parallel, delayed, cpu_count
import platform
from typing import Mapping


class DataFrameViewer(pandas.DataFrame):

	def __xml__(self):
		z = Elem('div')
		table = z.put('table', {'class':'dataframe larch_dataframe'})
		thead = table.put('thead')
		for i in range(self.index.nlevels):
			thead.put('th', text=self.index.names[i])
		for c in self.columns:
			thead.put('th', text=str(c))
		tbody = table.put('tbody')
		for i in self.index:
			row = tbody.put('tr')
			if self.index.nlevels==1:
				row.put('th', text=str(i))
			else:
				for ii in i:
					row.put('th', text=str(ii))
			for c in self.columns:
				cell = self.loc[i, c]
				try:
					cell_content = cell.__xml__()
				except AttributeError:
					if isinstance(cell, float):
						row.put('td', text="{0:.6g}".format(cell))
					else:
						row.put('td', text=str(cell))
				else:
					row.put('td') << cell_content
		return z

	def _repr_html_(self):
		return self.__xml__().tostring()

	def __getitem__(self, item):
		result = super().__getitem__(item)
		if type(result) == pandas.DataFrame:
			result.__class__ = DataFrameViewer
		return result

	def drop(self, *args, **kwargs):
		result = super().drop(*args, **kwargs)
		if type(result) == pandas.DataFrame:
			result.__class__ = DataFrameViewer
		return result


#####################
# DataFrame Styling
#####################

from matplotlib import colors
import seaborn

def global_background_gradient(s, m, M, cmap=None, low=0, high=0):
	if cmap is None:
		cmap = seaborn.light_palette("seagreen", as_cmap=True)
	rng = M - m
	norm = colors.Normalize(m - (rng * low),
							M + (rng * high))
	normed = norm(s.values)
	c = [colors.rgb2hex(x) for x in cmap(normed)]
	return ['background-color: %s' % color for color in c]



def apply_global_background_gradient(df, override_min=None, override_max=None, cmap=None, subset=None):
	if cmap is None:
		seagreen = seaborn.light_palette("seagreen", as_cmap=True)
		cmap = seagreen
	df = df.apply(
		global_background_gradient,
		cmap=cmap,
		m=override_min if override_min is not None else df.data.min().min(),
		M=override_max if override_max is not None else df.data.max().max(),
		subset=subset,
	)
	return df

###################

def columnize_with_joinable_backing(df, name, inplace=False, dtype=None, debug=False, backing=None):
	try:
		return columnize(df, name, inplace=inplace, dtype=dtype, debug=debug)
	except NameError:
		if not inplace:
			return columnize(df.join(backing), name, inplace=False, dtype=dtype, debug=debug)
		else:
			raise

def columnize_with_joinable_backing_2(df, name, dtype=None, debug=False, backing=None):
	source = df
	try:
		out = columnize(df, name, inplace=False, dtype=dtype, debug=debug)
	except NameError:
		source = df.join(backing)
		backing = None
		out = columnize(source, name, inplace=False, dtype=dtype, debug=debug)
	return out, source, backing

def columnize(df, name, inplace=True, dtype=None, debug=False, backing=None):
	"""Add a computed column to a DataFrame."""

	datanames = None
	from ..roles import LinearFunction # import here to prevent circular imports
	if isinstance(name, (list,tuple)):
		datanames = [str(_) for _ in name]
		datamap = {i:i for i in datanames}
	elif isinstance(name, Mapping):
		datanames = [str(_) for _ in name.values()]
		datamap = name
	elif isinstance(name, LinearFunction):
		datanames = [str(_.data) for _ in name]
		datamap = {i:i for i in datanames}

	if datanames is not None:
		if len(datanames) == 0:
			if inplace:
				return
			else:
				return pandas.DataFrame(index=df.index)
		columns = {}
		source = df
		for dname in set(datanames):
			out, source, backing = columnize_with_joinable_backing_2(source, dname, dtype, backing=backing)
			columns[dname] = out.to_numpy()
		df1 = pandas.DataFrame({k:columns[v] for (k,v) in datamap.items()}, index=df.index)
		if inplace:
			df[datamap.keys()] = df1.values
			return
		else:
			return df1

	# short circuit for natural names
	try:
		return df[name]
	except KeyError:
		pass

	from tokenize import tokenize, untokenize, NAME, OP, STRING, NUMBER
	from .aster import asterize
	DOT = (OP, '.')
	COLON = (OP, ':')
	COMMA = (OP, ',')
	OBRAC = (OP, '[')
	CBRAC = (OP, ']')
	OPAR = (OP, '(')
	CPAR = (OP, ')')
	EQUAL = (OP, '=')
	from io import BytesIO
	if inplace:
		recommand = [(NAME, 'df'), OBRAC, (STRING, f"'{name}'"), CBRAC, EQUAL]
	else:
		recommand = [ ]
	recommand += [OPAR,]
	try:
		name_encode = name.encode('utf-8')
	except AttributeError:
		name_encode = str(name).encode('utf-8')
	reader = BytesIO(name_encode)
	g = tokenize(reader.readline)
	from_df = False
	for toknum, tokval, _, _, _ in g:
		if toknum == NAME:
			pass
			if tokval in df.columns:
				# replace NAME tokens
				partial = [
					(NAME, 'df'),
					OBRAC, (STRING, f"'{tokval}'"), CBRAC,
				]
				from_df = True
				recommand.extend(partial)
			# break
			elif tokval in df.index.names:
				# replace NAME tokens with # df.index.get_level_values('NAME').values
				partial = [
					(NAME, 'df'), DOT, (NAME, 'index'), DOT, (NAME, 'get_level_values'),
					OPAR, (STRING, f"'{tokval}'"), CPAR, DOT, (NAME, 'values')
				]
				from_df = True
				recommand.extend(partial)
			else:
				# no dat contains this natural name
				# put the name back in raw, maybe it works cause it's a global, more likely
				# the exception manager below will catch it.
				recommand.append((toknum, tokval))
		else:
			recommand.append((toknum, tokval))
	if not from_df and not inplace:
		recommand = [(NAME, 'pandas'), DOT, (NAME, 'Series')] + recommand \
					+ [COMMA, (NAME, 'index'), EQUAL, (NAME, 'df'), DOT, (NAME, 'index')]
	recommand += [CPAR,]
	if dtype is not None:
		dtype_ = str(numpy.dtype(dtype))
		recommand += [DOT, (NAME, 'astype'), OPAR, (STRING, f"'{dtype_}'"), CPAR]
	try:
		ret = untokenize(recommand).decode('utf-8')
	except:
		print("<recommand>")
		print(recommand)
		print("</recommand>")
		raise
	g.close()
	reader.close()
	if debug:
		print("<ret>")
		print(ret)
		print("</ret>")
	j = asterize(ret, mode="exec" if inplace else "eval")
	from .aster import inXd
	from numpy import log, exp, log1p, absolute, fabs, sqrt, isnan, isfinite, logaddexp, fmin, fmax, nan_to_num, sin, cos, pi
	from ..util.common_functions import piece, normalize, boolean
	try:
		if inplace:
			_result = exec(j)
		else:
			_result = eval(j)
			if not isinstance(_result, pandas.Series):
				_result = pandas.Series(_result, index=df.index)
			try:
				_result.name = name
			except AttributeError:
				pass
	except Exception as exc:
		args = exc.args
		if not args:
			arg0 = ''
		else:
			arg0 = args[0]
		arg0 = arg0 + '\nwithin parsed command: "{!s}"'.format(name)
		if "max" in name:
			arg0 = arg0 + '\n(note to get the maximum of arrays use "fmax" not "max")'.format(name)
		if "min" in name:
			arg0 = arg0 + '\n(note to get the minimum of arrays use "fmin" not "min")'.format(name)
		if isinstance(exc, NameError):
			badname = str(exc).split("'")[1]
			goodnames = {
				'log', 'exp', 'log1p', 'absolute', 'fabs', 'sqrt', 'isnan',
				'isfinite', 'logaddexp', 'fmin', 'fmax', 'nan_to_num', 'piece', 'normalize',
			}
			goodnames |= set(df.columns)
			if backing is not None:
				goodnames |= set(backing.columns)
			from ..util.text_manip import case_insensitive_close_matches
			did_you_mean_list = case_insensitive_close_matches(badname, goodnames, n=3, cutoff=0.1, excpt=None)
			if len(did_you_mean_list) > 0:
				arg0 = arg0 + '\n' + "did you mean {}?".format(
					" or ".join("'{}'".format(s) for s in did_you_mean_list))
		exc.args = (arg0,) + args[1:]
		raise
	return _result

def compute(df, name, **kwargs):
	kwargs['inplace'] = False
	return columnize(df, name, **kwargs)


###################
### Typical Calibration Table

def counts_and_shares(
		tours_df,
		rows='tour_mode',
		cross='distH2Work',
		bins=None,
		stack=True,
		title=None,
		blurb=None,
		rowtotals=True,
):

	if isinstance(rows, str):
		if not isinstance(tours_df[rows].dtype, pandas.core.dtypes.dtypes.CategoricalDtype):
			raise TypeError("column for rows must be categorical")
		cats = tours_df[rows].copy()
	else:
		if not isinstance(rows.dtype, pandas.core.dtypes.dtypes.CategoricalDtype):
			raise TypeError("column for rows must be categorical")
		cats = rows.copy()

	firstcat, lastcat = cats.cat.categories[0], cats.cat.categories[-1]

	cats = cats.cat.add_categories(["TOTAL", "% of TOT"])

	if isinstance(cross, str):
		if bins=='cat':
			pivot_cols = pandas.Series(pandas.Categorical(tours_df[cross]), index=tours_df.index)
		elif bins is not None:
			pivot_cols = pandas.cut(
				tours_df[cross],
				bins=bins,
			)
		else:
			pivot_cols = tours_df[cross]
	elif isinstance(cross, pandas.Series):
		if isinstance(cross.dtype, pandas.core.dtypes.dtypes.CategoricalDtype):
			pivot_cols = cross
		else:
			raise TypeError('if a Series, cross must be a Categorical Series')
	else:
		raise TypeError(f'bad cross of type {type(cross)}')

	result1 = pandas.pivot_table(
		tours_df,
		values='personId',
		index=cats,
		columns=pivot_cols,
		aggfunc=len,
	).fillna(0).astype(int)

	result1.columns = result1.columns.astype(str)

	if rowtotals:
		result1.loc[:, 'TOTAL'] = result1.sum(1)

	result2 = (result1 / result1.sum(0))
	result2.loc['TOTAL', :] = numpy.nan
	result2.loc['% of TOT', :] = numpy.nan

	if rowtotals:
		result1.loc['TOTAL', :] = result1.sum(0)
		result1.loc['% of TOT', :] = -result1.loc['TOTAL', :] * 2 / result1.loc['TOTAL', :].sum()
	else:
		result1.loc['TOTAL', :] = result1.sum(0)
		result1.loc['% of TOT', :] = -result1.loc['TOTAL', :] / result1.loc['TOTAL', :].sum()

	def pct_or_thousands(x):
		if x < 0.0:
			return "{:.1%}".format(-x)
		else:
			return "{:,.0f}".format(x)

	def hide_nan_else_pct(x):
		if pandas.isnull(x):
			return ""
		else:
			return "{:.2%}".format(x)

	if stack:
		result = pandas.concat(
			[result1, result2],
			axis=1,
			keys=['Counts', 'Shares'],
			sort=False,
		)
		result = result.style.format({
			**{i: pct_or_thousands for i in result.columns if i[0] == 'Counts'},
			**{i: hide_nan_else_pct for i in result.columns if i[0] == 'Shares'}
		})

	else:
		result = (
			result1.style.format("{:,.0f}"),
			result2.style.format("{:.2%}"),
		)

	# try:
	# 	result = apply_global_background_gradient(
	# 		result,
	# 		override_max=0.75,
	# 		override_min=0.01,
	# 		# subset=[i for i in result.columns if i[0]=='Shares'],
	# 		subset=pandas.IndexSlice[firstcat:lastcat, [i for i in result.columns if i[0] == 'Shares']]
	# 	)
	# except ValueError:
	# 	pass

	result.applymap(
		lambda y: 'text-align:right;',
		# subset=pandas.IndexSlice[firstcat:lastcat, [i for i in result.columns if i[0]=='Counts']]
	)

	if title is None:
		return result
	else:
		from IPython.display import display, display_html, HTML
		display_html(HTML(f"<h2>{title}</h2>"))
		if blurb is not None:
			display_html(HTML(f"<p>{blurb}</p>"))
		display(result)





from .interface_info import ipython_status as _ipython_status
from .rate_limiter import NonBlockingRateLimiter
_message_set = _ipython_status()


class Throttlable():

	def __init__(self, throttle=None):
		if throttle is not None:
			self._throttle = NonBlockingRateLimiter(throttle)
		else:
			self._throttle = True



if 'IPython' in _message_set:
	from IPython.display import display, clear_output, display_html, HTML

	class display_head(Throttlable):

		def __init__(self, text, level=3, throttle=2):
			super().__init__(throttle)
			self.level = level
			self.tag = display(HTML(f'<h{self.level}>{text}</h{self.level}>'), display_id=True)

		def update(self, text, newline=False, force=False):
			if self._throttle or force:
				self.tag.update(HTML(f'<h{self.level}>{text}</h{self.level}>'))

		def __call__(self, *text, **kwargs):
			self.update(*text, **kwargs)

		def linefeed(self):
			pass

	class display_p(Throttlable):

		def __init__(self, text, force=False, throttle=2):
			super().__init__(throttle)
			if isinstance(text, str):
				self.tag = display(HTML(f'<p>{text}</p>'), display_id=True)
			else:
				self.tag = display(text, display_id=True)

		def update(self, text, force=False):
			if self._throttle or force:
				if isinstance(text, str):
					self.tag.update(HTML(f'<p>{text}</p>'))
				else:
					self.tag.update(text)

		def __call__(self, *text, **kwargs):
			self.update(*text, **kwargs)

else:

	class fake_display():

		def __init__(self, *args, **kwargs):
			print(*args)

		def update(self, *args):
			print(*args)

	display = fake_display
	clear_output = lambda *x,**y: None
	display_html = lambda *x,**y: None
	HTML = lambda *x,**y: None

	class display_head(Throttlable):

		def __init__(self, text, level=3, throttle=2):
			super().__init__(throttle)
			self.level = level
			print(f'{text}: ',end="")

		def update(self, text, newline=False, force=False):
			if self._throttle or force:
				if newline:
					print("")
					print(f'{text} ',end="")
				else:
					print(f'{text}: ', end="")

		def __call__(self, *text, **kwargs):
			self.update(*text, **kwargs)

		def linefeed(self):
			print("")

	class display_p(Throttlable):

		def __init__(self, text, force=False, throttle=2):
			super().__init__(throttle)
			if isinstance(text, str):
				print(text)
			else:
				if force:
					print(str(text))

		def update(self, text, force=False):
			if self._throttle or force:
				if isinstance(text, str):
					print(text)
				else:
					if force:
						print(str(text))

		def __call__(self, *text, **kwargs):
			self.update(*text, **kwargs)


class display_nothing():

	def __init__(self, *args, **kwargs):
		pass

	def update(self, *args, **kwargs):
		pass

	def __call__(self, *text, **kwargs):
		pass

	def linefeed(self):
		pass




from .idca import *
from ... import _reserved_names_
from ...general import selector_len_for, _sqz_same
from ...service import Pods

#
# class H5PodCS(H5PodCA):
# 	"""
# 	A HDF5 group node containing implicit :ref:`idca` format data created by stacking :ref:`idco` data.
#
# 	The CA pod is includes a selection of data arrays, where
# 	each array has rows of values attributable to casewise observations, and columns attributable to
# 	individual choice alternatives.  The ordering
# 	of the data is important, and all arrays in an H5PodCA must have the same shape
# 	and be ordered the same way, so that any group of arrays can be simply concantenated to
# 	create a 3-d :ref:`idca` array of data.  The same applies for multiple H5PodCA's that are
# 	associated together in a single H5Data object.
# 	"""
# 	def __init__(self, *args, **kwargs):
# 		super().__init__(*args, **kwargs)
# 		# if the _BUNCHES_ sub-group does not exist, create it if possible
# 		from ..h5util import get_or_create_subgroup
# 		b = get_or_create_subgroup(self._groupnode, '_BUNCHES_', skip_on_readonly=True)
# 		if b is None:
# 			self._bunches = {}
# 		else:
# 			from .. import H5Vault
# 			self._bunches = H5Vault(vaultnode=b)
# 		self._altcodes = []
#
# 	@property
# 	def podtype(self):
# 		return 'idcs'
#
# 	def set_bunch(self, bunch_name, bunch_def):
# 		if not isinstance(bunch_def, dict):
# 			raise TypeError('bunch_def must be a dict of alt:varname')
# 		self._bunches[bunch_name] = bunch_def
#
# 	def set_alts(self, alts):
# 		self._altcodes = alts
# 		return self
#
# 	def _get_list_from_bunchname(self, bunchname):
# 		if not isinstance(bunchname, str):
# 			return bunchname
# 		if bunchname not in self._bunches:
# 			raise KeyError(bunchname)
# 		names_dict = self._bunches[bunchname]
# 		result = []
# 		for n in self._altcodes:
# 			if n not in names_dict:
# 				result.append('0')
# 			else:
# 				result.append(names_dict[n])
# 		return result
#
# 	def __contains__(self, item):
# 		if item in self._bunches:
# 			return True
# 		return False
#
# 	def __getitem__(self, item):
# 		if len(self._altcodes) == 0:
# 			raise ValueError("alternatives vector not set")
#
# 		if isinstance(item, tuple) and len(item)>=2 and isinstance(item[-1], slice):
# 			names, slice_ = item[:-1], item[-1]
# 		else:
# 			names = item
# 			slice_ = None
#
# 		# convert a single name string to a one item list
# 		if isinstance(names, str):
# 			names = [names, ]
#
# 		names = [self._get_list_from_bunchname(i) for i in names]
#
# 		# check that all names are actually lists or tuples of the same length
# 		first_len = len(names[0])
# 		for cmd in names[1:]:
# 			assert( len(cmd) == first_len )
#
# 		def slice_len_for(slc, seqlen):
# 			if slc is None:
# 				return seqlen
# 			start, stop, step = slc.indices(seqlen)
# 			return max(0, (stop - start + (step - (1 if step > 0 else -1))) // step)
#
# 		dtype = numpy.float64
#
# 		result = numpy.zeros( [slice_len_for(slice_, self.shape[0]), *self.shape[1:-1], first_len, len(names)], dtype=dtype)
#
# 		for i, cmdgrp in enumerate(names):
# 			for j, cmd in enumerate(cmdgrp):
# 				result[...,j,i] = self._evaluate_single_item(cmd, slice_)
#
# 		return result
#
# 	def _load_into(self, names, slc, result):
# 		if len(self._altcodes) == 0:
# 			raise ValueError("alternatives vector not set")
# 		# convert a single name string to a one item list
# 		if isinstance(names, str):
# 			names = [names, ]
# 		names = [self._get_list_from_bunchname(i) for i in names]
#
# 		# check that all names are actually lists or tuples of the same length
# 		first_len = len(names[0])
# 		for cmd in names[1:]:
# 			assert( len(cmd) == first_len )
#
# 		def slice_len_for(slc, seqlen):
# 			if slc is None:
# 				return seqlen
# 			start, stop, step = slc.indices(seqlen)
# 			return max(0, (stop - start + (step - (1 if step > 0 else -1))) // step)
#
# 		assert (tuple(result.shape) == tuple([slice_len_for(slc, self.shape[0]), *self.shape[1:-1], first_len, len(names)]))
# 		for i, cmdgrp in enumerate(names):
# 			for j, cmd in enumerate(cmdgrp):
# 				result[...,j,i] = self._evaluate_single_item(cmd, slc)
# 		return result
#
# 	@property
# 	def shape(self):
# 		"""The shape of the pod.
#
# 		"""
# 		if 'SHAPE' in self._groupnode._v_attrs:
# 			return tuple(self._groupnode._v_attrs['SHAPE'][:]) + (len(self._altcodes),)
# 		if len(self.names()):
# 			for v in self._groupnode._v_children.values():
# 				try:
# 					found_shape = v.shape
# 				except:
# 					pass
# 				else:
# 					try:
# 						self.shape = found_shape
# 					except:
# 						pass
# 					return tuple(found_shape) + (len(self._altcodes),)
# 		raise NoKnownShape()
#
# 	def __xml__(self):
# 		from pprint import pformat
# 		x = super().__xml__()
# 		tr = x[0].elem('tr')
# 		tr.elem('th', text='bunches', style='vertical-align:top;')
# 		td = tr.elem('td')
# 		t1 = td.elem('table', cls='dictionary')
# 		header = t1.elem('thead').elem('tr')
# 		header.elem('th', text='name')
# 		header.elem('th', text='definition')
# 		for k,v in self._bunches.items():
# 			if k not in _reserved_names_:
# 				tr1 = t1.elem('tr')
# 				tr1.elem('td', text=k)
# 				tr1.elem('td', text=pformat(v))
# 		return x



class H5PodCS(H5PodCA):
	"""
	A HDF5 group node containing implicit :ref:`idca` format data created by stacking :ref:`idco` data.

	The CA pod is includes a selection of data arrays, where
	each array has rows of values attributable to casewise observations, and columns attributable to
	individual choice alternatives.  The ordering
	of the data is important, and all arrays in an H5PodCA must have the same shape
	and be ordered the same way, so that any group of arrays can be simply concantenated to
	create a 3-d :ref:`idca` array of data.  The same applies for multiple H5PodCA's that are
	associated together in a single H5Data object.
	"""
	def __init__(self, podgroup, storage=None, ident=None, alts=None, shape=None, **kwargs):
		if isinstance(podgroup, (list,tuple)) and not isinstance(podgroup, Pods):
			podgroup = Pods(podgroup)
		if not isinstance(podgroup, Pods):
			raise TypeError('podgroup must be Pods')
		super().__init__(do_nothing=True, ident=ident, shape=shape)
		self.podgroup = podgroup
		# if the _BUNCHES_ sub-group does not exist, create it if possible
		from ..h5util import get_or_create_subgroup
		b = get_or_create_subgroup(storage, '_BUNCHES_', skip_on_readonly=True)
		if b is None:
			self._bunches = {}
		else:
			from .. import H5Vault
			self._bunches = H5Vault(vaultnode=b)
		if alts is None:
			self._altcodes = []
		else:
			self._altcodes = alts
		self._groupnode = b
		for k,v in kwargs.items():
			self.set_bunch(k,v)

	@property
	def podtype(self):
		return 'idcs'

	def set_bunch(self, bunch_name, bunch_def):
		if not isinstance(bunch_def, dict):
			raise TypeError('bunch_def must be a dict of alt:varname')
		self._bunches[bunch_name] = bunch_def

	def set_alts(self, alts):
		self._altcodes = alts
		return self

	def _get_list_from_bunchname(self, bunchname):
		if not isinstance(bunchname, str):
			return bunchname
		if bunchname not in self._bunches:
			raise KeyError(bunchname)
		names_dict = self._bunches[bunchname]
		result = []
		for n in self._altcodes:
			if n not in names_dict:
				result.append('0')
			else:
				result.append(names_dict[n])
		return result

	def __contains__(self, item):
		if item in self._bunches:
			return True
		return False

	def names(self):
		return [i for i in self._bunches.keys() if i not in _reserved_names_]

	def __getitem__(self, item):
		if len(self._altcodes) == 0:
			raise ValueError("alternatives vector not set")

		if isinstance(item, tuple) and len(item)>=2 and isinstance(item[-1], slice):
			names, selector = item[:-1], item[-1]
		else:
			names = item
			selector = None

		# convert a single name string to a one item list
		if isinstance(names, str):
			names = [names, ]

		names = [self._get_list_from_bunchname(i) for i in names]

		# check that all names are actually lists or tuples of the same length
		first_len = len(names[0])
		for cmd in names[1:]:
			assert( len(cmd) == first_len )

		dtype = numpy.float64

		result = numpy.zeros( [selector_len_for(selector, self.shape[0]), *self.shape[1:-1], first_len, len(names)], dtype=dtype)

		for i, cmdgrp in enumerate(names):
			for j, cmd in enumerate(cmdgrp):
				# if selector is None:
				# 	result[...,j,i] = self.podgroup[ cmd ].squeeze()
				# else:
				# 	result[...,j,i] = self.podgroup[ cmd, selector ].squeeze()
				self.podgroup.load_data_item(cmd, result[..., j, i], selector)
		return result

	def load_data_item(self, name, result, selector=None):
		if len(self._altcodes) == 0:
			raise ValueError("alternatives vector not set")

		names = self._get_list_from_bunchname(name)

		required_shape = (selector_len_for(selector, self.shape[0]), *self.shape[1:-1], len(names))
		_sqz_same(result.shape, required_shape, comment=f'idcs for {name}')

		for i, cmd in enumerate(names):
			# if selector is None:
			# 	result[...,i] = self.podgroup[ cmd ].squeeze()
			# else:
			# 	# result[...,i] = self.podgroup[ cmd, selector ].squeeze()
			self.podgroup.load_data_item(cmd, result[..., i], selector)
		return result



	@property
	def shape(self):
		"""The shape of the pod.

		"""
		return tuple(self.podgroup.shape) + (len(self._altcodes),)

	def __xml__(self):
		from pprint import pformat
		x = super().__xml__(no_data=True)
		tr = x[0].elem('tr')
		tr.elem('th', text='bunches', style='vertical-align:top;')
		td = tr.elem('td')
		t1 = td.elem('table', cls='dictionary')
		header = t1.elem('thead').elem('tr')
		header.elem('th', text='name')
		header.elem('th', text='definition')
		for k,v in self._bunches.items():
			if k not in _reserved_names_:
				tr1 = t1.elem('tr')
				tr1.elem('td', text=k)
				tr1.elem('td', text=pformat(v))
		return x

	def __repr__(self):
		from pprint import pformat
		from ....util.text_manip import max_len
		s = f"<larch.{self.__class__.__name__}>"
		if self._groupnode is not None:
			s += f"\n |  file: {truncate_path_for_display(self.filename)}"
			if self._groupnode._v_pathname != "/":
				s += f"\n |  node: {self._groupnode._v_pathname}"
			try:
				s += f"\n |  shape: {self.shape}"
			except NoKnownShape:
				pass
		if len(self._bunches):
			s += "\n |  bunches:"
			for k,v in self._bunches.items():
				if k not in _reserved_names_:
					s += f"\n |    {k}: "
					s += f"\n |    {' '*len(k)}  ".join(pformat(v).split('\n'))
		else:
			s += "\n |  bunches: <empty>"
		return s


import pandas
import os
import gzip
import pickle
import base64
import io
import numpy
from pytest import approx
from larch.util.data_manipulation import periodize

def test_periodize():
	h = pandas.Series(range(21))
	p = periodize(h, default='OP', AM=(6.5, 9), PM=(16, 19))
	assert p.dtype == 'category'
	assert p[0] == 'OP'
	assert p[1] == 'OP'
	assert p[2] == 'OP'
	assert p[3] == 'OP'
	assert p[4] == 'OP'
	assert p[5] == 'OP'
	assert p[6] == 'OP'
	assert p[7] == 'AM'
	assert p[8] == 'AM'
	assert p[9] == 'AM'
	assert p[10] == 'OP'
	assert p[11] == 'OP'
	assert p[12] == 'OP'
	assert p[13] == 'OP'
	assert p[14] == 'OP'
	assert p[15] == 'OP'
	assert p[16] == 'PM'
	assert p[17] == 'PM'
	assert p[18] == 'PM'
	assert p[19] == 'PM'
	assert p[20] == 'OP'



#
# Adapted from the base pandas Styler.background_gradient
# https://github.com/pandas-dev/pandas/blob/v1.0.2/pandas/io/formats/style.py
#
# BSD 3-Clause License
#
# Copyright (c) 2008-2012, AQR Capital Management, LLC, Lambda Foundry, Inc. and PyData Development Team
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import numpy as np
import pandas as pd
from pandas.io.formats.style import _mpl, Styler
from typing import Optional
from pandas.core.dtypes.generic import ABCSeries
from pandas.core.indexes.api import Index
from pandas.api.types import is_dict_like, is_list_like


# the public IndexSlicerMaker
class _IndexSlice:
	"""
	Create an object to more easily perform multi-index slicing.

	See Also
	--------
	MultiIndex.remove_unused_levels : New MultiIndex with no unused levels.

	Notes
	-----
	See :ref:`Defined Levels <advanced.shown_levels>`
	for further info on slicing a MultiIndex.

	Examples
	--------
	>>> midx = pd.MultiIndex.from_product([['A0','A1'], ['B0','B1','B2','B3']])
	>>> columns = ['foo', 'bar']
	>>> dfmi = pd.DataFrame(np.arange(16).reshape((len(midx), len(columns))),
							index=midx, columns=columns)

	Using the default slice command:

	>>> dfmi.loc[(slice(None), slice('B0', 'B1')), :]
			   foo  bar
		A0 B0    0    1
		   B1    2    3
		A1 B0    8    9
		   B1   10   11

	Using the IndexSlice class for a more intuitive command:

	>>> idx = pd.IndexSlice
	>>> dfmi.loc[idx[:, 'B0':'B1'], :]
			   foo  bar
		A0 B0    0    1
		   B1    2    3
		A1 B0    8    9
		   B1   10   11
	"""

	def __getitem__(self, arg):
		return arg


IndexSlice = _IndexSlice()


def _non_reducing_slice(slice_):
	"""
	Ensure that a slice doesn't reduce to a Series or Scalar.

	Any user-passed `subset` should have this called on it
	to make sure we're always working with DataFrames.
	"""
	# default to column slice, like DataFrame
	# ['A', 'B'] -> IndexSlices[:, ['A', 'B']]
	kinds = (ABCSeries, np.ndarray, Index, list, str)
	if isinstance(slice_, kinds):
		slice_ = IndexSlice[:, slice_]

	def pred(part) -> bool:
		"""
		Returns
		-------
		bool
			True if slice does *not* reduce,
			False if `part` is a tuple.
		"""
		# true when slice does *not* reduce, False when part is a tuple,
		# i.e. MultiIndex slice
		return (isinstance(part, slice) or is_list_like(part)) and not isinstance(
			part, tuple
		)

	if not is_list_like(slice_):
		if not isinstance(slice_, slice):
			# a 1-d slice, like df.loc[1]
			slice_ = [[slice_]]
		else:
			# slice(a, b, c)
			slice_ = [slice_]  # to tuplize later
	else:
		slice_ = [part if pred(part) else [part] for part in slice_]
	return tuple(slice_)


def _maybe_numeric_slice(df, slice_, include_bool=False):
	"""
	Want nice defaults for background_gradient that don't break
	with non-numeric data. But if slice_ is passed go with that.
	"""
	if slice_ is None:
		dtypes = [np.number]
		if include_bool:
			dtypes.append(bool)
		slice_ = IndexSlice[:, df.select_dtypes(include=dtypes).columns]
	return slice_

class OverspecView(Styler):

	def __init__(self, data, *args, cmap='YlOrRd', high=0.3, **kwargs):
		if isinstance(data, (tuple,list)):
			data_ = pd.concat([
				pd.DataFrame(
					eigvec,
					index=eigpar,
					columns=[f"({n + 1}) {eigval:.4g}"]
				)
				for n, (eigval, eigpar, eigvec) in enumerate(data)
			], axis=1)
		elif data is None:
			data_ = pd.DataFrame()
		else:
			data_ = data
		super().__init__( data_, *args, **kwargs )
		self.absolute_background_gradient(cmap=cmap, high=high)

	def absolute_background_gradient(
			self,
			cmap="YlOrRd",
			low=0,
			high=0,
			axis=0,
			subset=None,
			text_color_threshold=0.408,
			vmin: Optional[float] = 0,
			vmax: Optional[float] = None,
	):
		"""
		Color the background in a gradient style,
		based on the magnitude (absolute value).
		The background color is determined according
		to the data in each column (optionally row).
		Requires matplotlib.


		Parameters
		----------
		cmap : str or colormap
			Matplotlib colormap.
		low : float
			Compress the range by the low.
		high : float
			Compress the range by the high.
		axis : {0 or 'index', 1 or 'columns', None}, default 0
			Apply to each column (``axis=0`` or ``'index'``), to each row
			(``axis=1`` or ``'columns'``), or to the entire DataFrame at once
			with ``axis=None``.
		subset : IndexSlice
			A valid slice for ``data`` to limit the style application to.
		text_color_threshold : float or int
			Luminance threshold for determining text color. Facilitates text
			visibility across varying background colors. From 0 to 1.
			0 = all text is dark colored, 1 = all text is light colored.
		vmin : float, default 0
			Minimum data value that corresponds to colormap minimum value.
			When None, the minimum value of the data will be used.
		vmax : float, optional
			Maximum data value that corresponds to colormap maximum value.
			When None (default): the maximum value of the data will be used.
		Returns
		-------
		self : Styler
		Raises
		------
		ValueError
			If ``text_color_threshold`` is not a value from 0 to 1.
		Notes
		-----
		Set ``text_color_threshold`` or tune ``low`` and ``high`` to keep the
		text legible by not using the entire range of the color map. The range
		of the data is extended by ``low * (x.max() - x.min())`` and ``high *
		(x.max() - x.min())`` before normalizing.
		"""
		subset = _maybe_numeric_slice(self.data, subset)
		subset = _non_reducing_slice(subset)
		self.apply(
			self._absolute_background_gradient,
			cmap=cmap,
			subset=subset,
			axis=axis,
			low=low,
			high=high,
			text_color_threshold=text_color_threshold,
			vmin=vmin,
			vmax=vmax,
		)
		return self

	@staticmethod
	def _absolute_background_gradient(
			s,
			cmap="YlOrRd",
			low=0,
			high=0,
			text_color_threshold=0.408,
			vmin: Optional[float] = 0,
			vmax: Optional[float] = None,
	):
		"""
		Color background in a range according to the data.
		"""
		if (
				not isinstance(text_color_threshold, (float, int))
				or not 0 <= text_color_threshold <= 1
		):
			msg = "`text_color_threshold` must be a value from 0 to 1."
			raise ValueError(msg)

		with _mpl(Styler.background_gradient) as (plt, colors):
			smin = np.nanmin(np.absolute(s.to_numpy())) if vmin is None else vmin
			smax = np.nanmax(np.absolute(s.to_numpy())) if vmax is None else vmax
			rng = smax - smin
			# extend lower / upper bounds, compresses color range
			norm = colors.Normalize(smin - (rng * low), smax + (rng * high))
			# matplotlib colors.Normalize modifies inplace?
			# https://github.com/matplotlib/matplotlib/issues/5427
			rgbas = plt.cm.get_cmap(cmap)(norm(np.absolute(s.to_numpy(dtype=float))))

			def relative_luminance(rgba):
				"""
				Calculate relative luminance of a color.
				The calculation adheres to the W3C standards
				(https://www.w3.org/WAI/GL/wiki/Relative_luminance)
				Parameters
				----------
				color : rgb or rgba tuple
				Returns
				-------
				float
					The relative luminance as a value from 0 to 1
				"""
				r, g, b = (
					x / 12.92 if x <= 0.03928 else ((x + 0.055) / 1.055 ** 2.4)
					for x in rgba[:3]
				)
				return 0.2126 * r + 0.7152 * g + 0.0722 * b

			def css(rgba):
				dark = relative_luminance(rgba) < text_color_threshold
				text_color = "#f1f1f1" if dark else "#000000"
				return f"background-color: {colors.rgb2hex(rgba)};color: {text_color};"

			if s.ndim == 1:
				return [css(rgba) for rgba in rgbas]
			else:
				return pd.DataFrame(
					[[css(rgba) for rgba in row] for row in rgbas],
					index=s.index,
					columns=s.columns,
				)



import numpy
import pandas
import ast
import re


### CATEGORICAL / ONE-HOT EXPANSION ###

def _to_categorical(y, num_classes=None, dtype='float32'):
	"""Converts a class vector (integers) to binary class matrix.

	E.g. for use with categorical_crossentropy.

	Parameters
	----------
	y: class vector to be converted into a matrix
		(integers from 0 to num_classes).
	num_classes: total number of classes.
	dtype: The data type expected by the input, as a string
		(`float32`, `float64`, `int32`...)

	Returns
	-------
	A binary matrix representation of the input. The classes axis
		is placed last.

	Notes
	-----
	This function is from keras.utils
	"""
	y = numpy.array(y, dtype='int')
	input_shape = y.shape
	if input_shape and input_shape[-1] == 1 and len(input_shape) > 1:
		input_shape = tuple(input_shape[:-1])
	y = y.ravel()
	if not num_classes:
		num_classes = numpy.max(y) + 1
	n = y.shape[0]
	categorical = numpy.zeros((n, num_classes), dtype=dtype)
	categorical[numpy.arange(n), y] = 1
	output_shape = input_shape + (num_classes,)
	categorical = numpy.reshape(categorical, output_shape)
	return categorical

def to_categorical(arr, index=None, dtype='float32'):

	if isinstance(arr, (pandas.DataFrame, pandas.Series)) and index is None:
		index = arr.index

	codes, positions = numpy.unique(arr, return_inverse=True)
	return pandas.DataFrame(
		data=_to_categorical(positions, dtype=dtype),
		columns=codes,
		index=index,
	)

def categorical_expansion(s, column=None, inplace=False, drop=False):
	"""
	Expand a pandas Series into a DataFrame containing a categorical dummy variables.

	This is sometimes called "one-hot" encoding.

	Parameters
	----------
	s : pandas.Series or pandas.DataFrame
		The input data
	column : str, optional
		If `s` is given as a DataFrame, expand this column
	inplace : bool, default False
		If true, add the result directly as new columns in `s`.
	drop : bool, default False
		If true, drop the existing column from `s`. Has no effect if
		`inplace` is not true.

	Returns
	-------
	pandas.DataFrame
		Only if inplace is false.
	"""
	if isinstance(s, pandas.DataFrame):
		input = s
		if column is not None and column not in s.columns:
			raise KeyError(f'key not found "{column}"')
		if len(s.columns) == 1:
			s = s.iloc[:, 0]
			column = s.columns[0]
		elif column is not None:
			s = s.loc[:, column]
	else:
		input = None

	onehot = to_categorical(s)
	onehot.columns = [f'{s.name}=={_}' for _ in onehot.columns]
	if inplace and input is not None:
		input[onehot.columns] = onehot
		if drop:
			input.drop([column], axis=1)
	else:
		return onehot

one_hot_expansion = categorical_expansion

def categorical_compression(df, columns=None, inplace=False, drop=False, name=None):
	"""
	Compress a one-hot encoded part of a pandas DataFrame into a categorical variable.

	Parameters
	----------
	s : pandas.DataFrame
		The input data
	columns : Sequence[str], optional
		Use these columns of `df`, or all columns if not given.
	inplace : bool, default False
		If true, add the result directly as new columns in `df`.
	drop : bool, default False
		If true, drop the existing columns from `df`. Has no effect if
		`inplace` is not true.

	Returns
	-------
	pandas.Series
		Only if inplace is false.
	"""
	input = df
	if columns is None:
		columns = df.columns
	onehot = df[columns]
	result = pandas.Series(
		numpy.full(onehot.index.shape, numpy.nan),
		dtype=pandas.CategoricalDtype(columns),
		index=onehot.index,
	)
	for i in onehot.columns:
		result.loc[onehot[i]==1] = i
	if inplace and name is not None:
		input[name] = result
		if drop:
			input.drop(columns, axis=1)
	else:
		return result

one_hot_compression = categorical_compression


### FOURIER EXPANSION ###

def fourier_expansion_names(basename, length=4):
	"""
	Get the names of items for a fourier series expansion.

	Parameters
	----------
	basename : str
		The input data name
	length : int
		Length of expansion series

	Returns
	-------
	list of str
	"""
	columns = []
	for i in range(length):
		func = 'cos' if i % 2 else 'sin'
		mult = ((i // 2) + 1) * 2
		columns.append(f'{func}({basename}*{mult}*pi)')
	return columns


def fourier_expansion(s, length=4, column=None, inplace=False):
	"""
	Expand a pandas Series into a DataFrame containing a fourier series.

	Parameters
	----------
	s : pandas.Series or pandas.DataFrame
		The input data
	length : int
		Length of expansion series
	column : str, optional
		If `s` is given as a DataFrame, use this column

	Returns
	-------
	pandas.DataFrame
	"""
	if isinstance(s, pandas.DataFrame):
		input = s
		if len(s.columns) == 1:
			s = s.iloc[:, 0]
		elif column is not None:
			s = s.loc[:, column]
	else:
		input = None
	columns = fourier_expansion_names(s.name, length=length)
	df = pandas.DataFrame(
		data=0,
		index=s.index,
		columns=columns,
	)
	for i, col in enumerate(columns):
		func = numpy.cos if i % 2 else numpy.sin
		mult = ((i // 2) + 1) * 2
		df.iloc[:, i] = func(s * mult * numpy.pi)
	if inplace and input is not None:
		input[columns] = df
	else:
		return df


### PIECEWISE LINEAR ###

def piece(x, low_bound, high_bound):
	"""
	Clip the values in an array.

	This function differs from the usual `numpy.clip`
	in that the result is shifted by `low_bound` if it is
	given, so that the valid result range is a contiguous
	block of non-negative values starting from 0.

	Parameters
	----------
	x : array-like
		Array containing elements to clip
	low_bound, high_bound : scalar, array-like, or None
		Minimum and maximum values. If None, clipping is not
		performed on the corresponding edge. Both may be
		None, which results in a noop. Both are broadcast
		against `x`.

	Returns
	-------
	clipped_array : array-like
	"""
	if low_bound is None:
		if high_bound is None:
			return x
		else:
			return numpy.minimum(x, high_bound)
	else:
		if high_bound is None:
			return numpy.maximum(x-low_bound, 0)
		else:
			return numpy.clip(x, low_bound, high_bound)-low_bound


def parse_piece(s):
	if s is not None:
		float_regex = '[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?'
		z = re.match(f"^piece\((.*),({float_regex}|None),({float_regex}|None)\)$", s)
		if z:
			# data is exactly and only a piecewise part
			base_def = z.group(1)
			lowbound = ast.literal_eval(z.group(2))
			highbound = ast.literal_eval(z.group(4))
			return base_def, lowbound, highbound, True
		z = re.match(f".*piece\((.*),({float_regex}|None),({float_regex}|None)\)", s)
		if z:
			# data is a piecewise part possibly interacted with something else
			base_def = z.group(1)
			lowbound = ast.literal_eval(z.group(2))
			highbound = ast.literal_eval(z.group(4))
			return base_def, lowbound, highbound, False
	return None, None, None, None

def piecewise_linear(x, p=None, breaks=None):
	from ..roles import P,X
	from ..model import linear

	if p is None:
		p = x

	# Flip x and p if given backwards
	if isinstance(x, linear.ParameterRef_C) and isinstance(p, linear.DataRef_C):
		p,x = x,p

	if breaks is None:
		raise ValueError('missing required argument `breaks`')

	xs = piecewise_linear_data_names(x, breaks)
	ps = piecewise_linear_parameter_names(p, breaks)

	return sum(X(x_) * P(p_) for x_,p_ in zip(xs,ps))


def piecewise_linear_parameter_names(basename, breaks):
	from ..roles import P,X

	lex = '②③④⑤⑥⑦⑧⑨⑩⑪⑫⑬⑭⑮⑯⑰⑱⑲⑳'

	# first leg
	b = breaks[0]
	z = [P(f"{basename} ① up to {b}")]
	# middle legs
	for i in range(len(breaks) - 1):
		b0 = breaks[i]
		b1 = breaks[i + 1]
		z += [P(f"{basename} {lex[i]} {b0} to {b1}")]
	# last leg
	i = len(breaks) - 1
	b0 = breaks[i]
	z += [P(f"{basename} {lex[i]} over {b0}")]
	return z


def piecewise_linear_data_names(basename, breaks):
	# first leg
	b = breaks[0]
	z = [f"piece({basename},None,{b})"]
	# middle legs
	for i in range(len(breaks) - 1):
		b0 = breaks[i]
		b1 = breaks[i + 1]
		z += [f"piece({basename},{b0},{b1})"]
	# last leg
	i = len(breaks) - 1
	b0 = breaks[i]
	z += [f"piece({basename},{b0},None)"]
	return z


def infer_breaks(fun):
	"""
	Infer break points from a piecewise_linear LinearFunction.

	Parameters
	----------
	fun : LinearFunction_C
	"""
	m = re.compile('piece\(.*,(.*),(.*)\)')
	_0 = m.match(fun[0].data)
	if _0.group(1) != 'None':
		raise NotImplementedError('Bounded Minimum')
	breaks = [ast.literal_eval(_0.group(2))]
	for i in fun[1:-1]:
		_n = m.match(i.data)
		if breaks[-1] != ast.literal_eval(_n.group(1)):
			raise NotImplementedError('Diconnected')
		breaks.append(ast.literal_eval(_n.group(2)))
	_n = m.match(fun[-1].data)
	if breaks[-1] != ast.literal_eval(_n.group(1)):
		raise NotImplementedError('Diconnected Top')
	if _n.group(2) != 'None':
		raise NotImplementedError('Bounded Maximum')
	return breaks

def piecewise_expansion(s, breaks, column=None, inplace=False):
	"""
	Expand a pandas Series into a DataFrame containing a piecewise linear series.

	Parameters
	----------
	s : pandas.Series or pandas.DataFrame
		The input data
	breaks : list
		Piecewise linear articulation breakpoints
	column : str, optional
		If `s` is given as a DataFrame, use this column

	Returns
	-------
	pandas.DataFrame
	"""
	if isinstance(s, pandas.DataFrame):
		input = s
		if len(s.columns) == 1:
			s = s.iloc[:, 0]
		elif column is not None:
			s = s.loc[:, column]
	else:
		input = None

	from ..model import linear

	if isinstance(breaks, linear.LinearFunction_C):
		breaks = infer_breaks(breaks)

	columns = piecewise_linear_data_names(s.name, breaks=breaks)
	df = pandas.DataFrame(
		data=0,
		index=s.index,
		columns=columns,
	)

	# first leg
	b = breaks[0]
	df.iloc[:, 0] = piece(s, None, b)

	# middle legs
	for i in range(len(breaks) - 1):
		b0 = breaks[i]
		b1 = breaks[i + 1]
		df.iloc[:, i + 1] = piece(s, b0, b1)
	# last leg
	i = len(breaks) - 1
	b0 = breaks[i]
	df.iloc[:, i + 1] = piece(s, b0, None)

	if inplace and input is not None:
		input[columns] = df
	else:
		return df


### IDCE MANIPULATIONS ###

def _nonzero_minimum_rescale_slower(x):
	x_ = x[x!=0]
	if x_.size == 0:
		s = 1
	else:
		s = numpy.min(x[x!=0])
	return x/s

def _nonzero_minimum_rescale(x):
	s = numpy.ma.masked_equal(x, 0.0, copy=False).min()
	if numpy.isnan(s) or s==0:
		s = 1
	return x/s


def ratio_min(s, column=None, inplace=False, result_name=None):
	"""
	Modify a pandas Series or DataFrame in `idce` format to scale within cases relative to the minimum value.

	Parameters
	----------
	s : pandas.Series or pandas.DataFrame
		The input data
	column : str, optional
		If `s` is given as a DataFrame, use this column
	inplace : bool, default False
		modify `s` in-place?
	result_name : str, optional
		If `s` is given as a DataFrame, and `inplace` is True, write to this column.
		When not given, the result is named "ratio_min(<column>)"

	Returns
	-------
	pandas.DataFrame

	Raises
	------
	ValueError
		The input argument `s` is not in idce format (i.e., it does not have
		a MultiIndex as the index or that MultiIndex does not have 2 levels.)
	"""
	index = s.index

	if not isinstance(s.index, pandas.MultiIndex) or len(s.index.levels) != 2:
		raise ValueError('`s` must be in idce format, with a 2-level multi-index.')

	if column is None and isinstance(s, pandas.Series):
		column = s.name

	if isinstance(s, pandas.DataFrame):
		input = s
		if len(s.columns) == 1:
			s = s.iloc[:, 0]
		elif column is not None:
			s = s.loc[:, column]
	else:
		input = None

	divisor = s.groupby(index.codes[0]).transform(numpy.min)
	divisor[divisor==0] = 1.0
	scaled = s/divisor

	# scaled = s.groupby(index.labels[0]).transform(_nonzero_minimum_rescale)

	if result_name is None and column is not None:
		result_name = f'ratio_min({column})'

	if inplace and input is not None and result_name is not None:
		input[result_name] = scaled
	else:
		return scaled



import numpy
import pandas
import enum
from .. import warning
from .histograms import sizable_histogram_figure, seems_like_discrete_data
from . import dictx
from .arraytools import scalarize
import numpy.ma as ma
from .common_functions import parse_piece


def statistics_for_array(
		arr,
		histogram=True,
		varname=None,
		ch_weights=None,
		avail=None,
		dictionary=None,
		flatten=False,
		on_error='ignore',
		**kwargs,
):
	"""
	Generate a statistical analysis of an array.

	Parameters
	----------
	arr : ndarray
		The array to analyze.
	histogram : bool or numeric
		If this evaluates as True, a histogram for the data array
		will be generated. If given as a number, this indicates the
		size and level of detail in the histogram.
	varname : str, optional
		A label for the x-axis of the histogram.  Only appears if
		the size is 3 or greater.  The varname is also used to
		detect piecewise linear segments, when given as
		"piece(label, lowerbound, upperbound)". Note this function
		does not *process* the array to extract a linear segment,
		this just flags the histogram generator to know that this
		is a linear segment.
	ch_weights : ndarray, optional
		If given, this indicates the choice-weights of the observations.
		Must be the same shape as `arr`. Useful to analyze whether the
		distribution of the data appears different for chosen
		alternatives than for all alternatives.
	avail : ndarray, optional
		If given, this indicates the availability of the observations.
		Must be the same shape as `arr`. If missing, all finite values
		in `arr` are considered available.
	dictionary : dict, optional
		A dictionary of {value: str} pairs, which give labels to use on
		the histogram instead of the code values stored in `arr`.
	flatten : bool, default False
		Indicates if `arr` (and `ch_weights` if given) should be
		flattened before processing.
	on_error : {'ignore', 'raise'}
		How to handle exceptions encountered while generating statistics.

	Other keyword parameters are passed through to
	:meth:`sizable_histogram_figure`.

	Returns
	-------
	dictx

	"""
	a = arr

	if flatten:
		a = a.reshape(-1)
		if ch_weights is not None:
			ch_weights = ch_weights.reshape(-1)

	stats = dictx()

	if varname is not None:
		stats.description = varname

	base_def, lowbound, highbound, piece_exactly = parse_piece(varname)

	try:
		nans = numpy.isnan(a)
	except TypeError:
		can_nan = False
	else:
		can_nan = True

	a_masked_is_category = None

	if avail is None and can_nan:
		a_masked = ma.masked_array(a, mask=~numpy.isfinite(a))
		stats.n = scalarize(numpy.isfinite(a).sum(axis=0))
	elif avail is not None:
		a_masked = ma.masked_array(a, mask=~avail)
		stats.n = scalarize((avail!=0).sum(axis=0))
	elif isinstance(a, pandas.Series) and isinstance(a.dtype, pandas.CategoricalDtype):
		a_masked = ma.masked_array(a, mask=pandas.isnull(a))
		a_masked_is_category = True
		stats.n = scalarize((~pandas.isnull(a)).sum(axis=0))
	else:
		a_masked = a
		stats.n = scalarize(a.shape[0])

	if can_nan:
		stats.minimum = scalarize(numpy.nanmin(a_masked, axis=0))
		stats.maximum = scalarize(numpy.nanmax(a_masked, axis=0))
		with warning.ignore_warnings():
			if ma.is_masked:
				stats.median = scalarize(ma.median(a_masked, axis=0))
			else:
				stats.median = scalarize(numpy.nanmedian(a_masked, axis=0))

	if ch_weights is not None:
		a_ch_weighted = ma.masked_array(a, mask=(ch_weights<=0))
		stats.n_chosen = scalarize(a_ch_weighted.count(axis=0))
		if can_nan:
			stats.minimum_chosen = scalarize(numpy.nanmin(a_ch_weighted, axis=0))
			stats.maximum_chosen = scalarize(numpy.nanmax(a_ch_weighted, axis=0))
			with warning.ignore_warnings():
				stats.median_chosen = scalarize(numpy.nanmedian(a_ch_weighted, axis=0))

	if ch_weights is not None:
		if avail is None and can_nan:
			ch_weightsx = ma.masked_array(ch_weights, mask=~numpy.isfinite(a))
		elif avail is not None:
			ch_weightsx = ma.masked_array(ch_weights, mask=~avail)
		elif isinstance(a, pandas.Series) and isinstance(a.dtype, pandas.CategoricalDtype):
			ch_weightsx = ma.masked_array(ch_weights, mask=pandas.isnull(a))
		else:
			ch_weightsx = ch_weights
	else:
		ch_weightsx = None

	if a_masked_is_category is None:
		try:
			a_masked_is_category = isinstance(a_masked.dtype, pandas.CategoricalDtype)
		except:
			a_masked_is_category = False

	if (can_nan and histogram) or a_masked_is_category:
		with warning.ignore_warnings():

			discrete_ = kwargs.pop('discrete', None)
			try:
				if a_masked_is_category:
					discrete_ = True
			except TypeError:
				pass
			if discrete_ is None:
				discrete_ = seems_like_discrete_data(a_masked, dictionary)

			if lowbound is not None or highbound is not None:
				stats.histogram = sizable_histogram_figure(
					a_masked, sizer=histogram,
					title=None, xlabel=base_def, ylabel='Frequency',
					ch_weights=ch_weightsx,
					piecerange=(lowbound, highbound),
					**kwargs
				)
			else:
				stats.histogram = sizable_histogram_figure(
					a_masked, sizer=histogram,
					title=None, xlabel=varname, ylabel='Frequency',
					ch_weights=ch_weightsx,
					discrete=discrete_,
					on_error=on_error,
					**kwargs
				)
	if can_nan:
		stats.mean = scalarize(numpy.mean(a_masked, axis=0))
		stats.stdev = scalarize(numpy.std(a_masked, axis=0))
		stats.zeros = scalarize(numpy.sum(numpy.logical_not(a_masked), axis=0))
		with warning.ignore_warnings():
			stats.positives = scalarize(numpy.sum(a_masked > 0, axis=0))
			stats.negatives = scalarize(numpy.sum(a_masked < 0, axis=0))
		n_nans = scalarize(numpy.sum(numpy.isnan(a), axis=0))
		if numpy.any(n_nans):
			stats.nans = n_nans
		n_infs = scalarize(numpy.sum(numpy.isinf(a), axis=0))
		if numpy.any(n_infs):
			stats.infs = n_infs
		a_masked.mask |= ( a_masked ==0)
		stats.nonzero_minimum = scalarize(numpy.min(a_masked, axis=0))
		stats.nonzero_maximum = scalarize(numpy.max(a_masked, axis=0))
		stats.nonzero_mean = scalarize(numpy.mean(a_masked, axis=0))
		stats.nonzero_stdev = scalarize(numpy.std(a_masked, axis=0))
	elif a_masked_is_category:
		stats.nulls = pandas.isnull(a).sum()
		the_mode = tuple(a.mode().values)
		if len(the_mode) == 1:
			the_mode = the_mode[0]
		stats.mode = the_mode

	if dictionary is not None:
		stats.dictionary = dictionary

	return stats

def statistics_for_array5(arr, histogram=5, *args, **kwargs):
	"""
	Generate a statistical analysis of an array.

	Parameters
	----------
	arr : ndarray
		The array to analyze.
	histogram : bool or numeric, default 5
		If this evaluates as True, a histogram for the data array
		will be generated. If given as a number, this indicates
		the size and level of detail in the histogram.
	varname : str, optional
		A label for the x-axis of the histogram.  Only appears if
		the size is 3 or greater.  The varname is also used to
		detect piecewise linear segments, when given as
		"piece(label, lowerbound, upperbound)". Note this function
		does not *process* the array to extract a linear segment,
		this just flags the histogram generator to know that this
		is a linear segment.
	ch_weights : ndarray, optional
		If given, this indicates the choice-weights of the
		observations.  Must be the same shape as `arr`. Useful to
		analyze whether the distribution of the data appears
		different for chosen alternatives than for all alternatives.
	avail : ndarray, optional
		If given, this indicates the availability of the observations.
		Must be the same shape as `arr`. If missing, all finite values
		in `arr` are considered available.
	dictionary : dict, optional
		A dictionary of {value: str} pairs, which give labels to use
		on the histogram instead of the code values stored in `arr`.
	flatten : bool, default False
		Indicates if `arr` (and `ch_weights` if given) should be
		flattened before processing.

	Other keyword parameters are passed through to
	:meth:`sizable_histogram_figure`.

	Returns
	-------
	dictx

	"""
	return statistics_for_array(arr, histogram=histogram, *args, **kwargs)

def statistics_for_dataframe(df, histogram=True, ch_weights=None, avail=None, **kwargs):
	s = {}
	for col in df.columns:
		if isinstance(df.dtypes[col], pandas.CategoricalDtype):
			s[col] = statistics_for_array(
				df[col],
				varname=col,
				histogram=histogram,
				ch_weights=ch_weights,
				avail=avail,
				**kwargs,
			)
		elif numpy.issubdtype(df.dtypes[col], numpy.floating):
			s[col] = statistics_for_array(
				df[col],
				varname=col,
				histogram=histogram,
				ch_weights=ch_weights,
				avail=avail,
				**kwargs,
			)
		elif numpy.issubdtype(df.dtypes[col], numpy.integer):
			s[col] = statistics_for_array(
				df[col],
				varname=col,
				histogram=histogram,
				ch_weights=ch_weights,
				avail=avail,
				**kwargs,
			)
		else:
			s[col] = statistics_for_array(
				df[col],
				varname=col,
				histogram=histogram,
				ch_weights=ch_weights,
				avail=avail,
				**kwargs,
			)

	from .dataframe import DataFrameViewer
	result = DataFrameViewer.from_dict(s, orient='index')
	if 'nans' in result.columns:
		result['nans'].fillna(0, inplace=True)
	try:
		result = result.drop('description', axis=1)
	except:
		pass
	return result


def uniques(s, dictionary=None):
	"""
	Count the frequency for each unique value in a pandas.Series.

	Parameters
	----------
	s : pandas.Series
		The values to count.
	dictionary : Mapping or Enum, optional
		A mapping converting the actual values in the series
		to a more meaningful value (e.g. changes code numbers
		to names).  If an Enum class is passed, the values
		are treated as keys, and the names as values.

	Returns
	-------
	pandas.Series
	"""
	action = s
	len_action = len(action)
	try:
		action = action[~numpy.isnan(action)]
	except TypeError:
		num_nan = 0
	else:
		num_nan = len_action - len(action)
	x = numpy.unique(action, return_counts=True)
	if dictionary is None:
		y = pandas.Series(x[1], x[0])
	else:
		if isinstance(dictionary, enum.EnumMeta):
			dictionary = {v: k for k, v in dictionary.__members__.items()}
		y = pandas.Series(x[1], [dictionary.get(j, j) for j in x[0]])
	if num_nan:
		y[numpy.nan] = num_nan
	return y


def invmap(s, mapping):
	"""
	Apply an inverse map.

	When the values in `s` are the values of a Mapping,
	but you want them to be the keys instead.  Note
	that values need not be unique, so an arbitrary
	key with the correct value will result for non-unique
	inverse mappings.

	Parameters
	----------
	s : array-like
	mapping : Mapping or Enum, optional
		A mapping. If an Enum class is passed, the
		__members__ is used.

	Returns
	-------

	"""
	if isinstance(mapping, enum.EnumMeta):
		mapping = mapping.__members__
	inv_map = {v: k for k, v in mapping.items()}
	return s.map(inv_map)



from io import BytesIO
from xmle import Elem
import pandas
import numpy
from typing import Collection, Callable, Mapping

def adjust_spines(ax, spines):
	for loc, spine in ax.spines.items():
		if loc in spines:
			#spine.set_position(('outward', 10))  # outward by 10 points
			#spine.set_smart_bounds(True)
			pass
		else:
			spine.set_color('none')  # don't draw spine
	# turn off ticks where there is no spine
	if 'left' in spines:
		ax.yaxis.set_ticks_position('left')
	else:
		# no yaxis ticks
		ax.yaxis.set_ticks([])
	if 'bottom' in spines:
		ax.xaxis.set_ticks_position('bottom')
	else:
		# no xaxis ticks
		ax.xaxis.set_ticks([])



def plot_as_svg_xhtml(
		fig,
		classname='figure',
		headerlevel=2,
		header=None,
		anchor=1,
		transparent=True,
		tooltip=None,
		bbox_extra_artists=None,
		pad_inches=0.1,
		**format,
):
	existing_format_keys = list(format.keys())
	for key in existing_format_keys:
		if key.upper()!=key: format[key.upper()] = format[key]
	if 'GRAPHWIDTH' not in format and 'GRAPHHEIGHT' in format: format['GRAPHWIDTH'] = format['GRAPHHEIGHT']
	if 'GRAPHWIDTH' in format and 'GRAPHHEIGHT' not in format: format['GRAPHHEIGHT'] = format['GRAPHWIDTH']*.67
	import xml.etree.ElementTree as ET
	ET.register_namespace("","http://www.w3.org/2000/svg")
	ET.register_namespace("xlink","http://www.w3.org/1999/xlink")
	imgbuffer = BytesIO()
	fig.savefig(
		imgbuffer, dpi=None, facecolor="none", edgecolor='w',
		orientation='portrait', format='svg',
		transparent=transparent, bbox_inches="tight", pad_inches=pad_inches,
		bbox_extra_artists=bbox_extra_artists,
	)
	x = Elem("div", {'class':classname})
	if header:
		x.hn(headerlevel, header, anchor=anchor)
	x << ET.fromstring(imgbuffer.getvalue().decode())
	if tooltip is not None:
		x[0][1].insert(0, Elem("title", text=tooltip))
	return x



def line_graph(
		x,
		y,
		x_title=None,
		y_title=None,
		show_legend=None,
		**kwargs,
):
	"""
	Generate a line graph.

	Parameters
	----------
	x : array-like
	y : array-like or pandas.Series or pandas.DataFrame
	x_title : str
	y_title : str
	show_legend : bool, optional
	**kwargs
	"""
	useful_legend = False

	if x_title is None:
		x_title = getattr(x, 'name', None)

	from matplotlib import pyplot as plt
	fig, ax = plt.subplots()

	if isinstance(y, pandas.DataFrame):
		for col in y.columns:
			ax.plot(x, y[col], label=col)
			useful_legend = True
	elif isinstance(y, pandas.Series):
		plt.plot(x, y.values, label=y.name)
		if y.name is not None:
			useful_legend = True
	elif isinstance(y, Collection) and all(isinstance(k, Callable) for k in y):
		for k in y:
			this_y = k(x)
			label_y = getattr(this_y, 'name', None)
			ax.plot(x, this_y, label=label_y)
			useful_legend = useful_legend or (label_y is not None)
	elif isinstance(y, Mapping):
		for col, val in y.items():
			ax.plot(x, val, label=col)
		useful_legend = True
	else:
		ax.plot(x, y)

	if x_title is not None:
		ax.set_xlabel(x_title)

	if y_title is not None:
		ax.set_ylabel(y_title)

	if show_legend is None and useful_legend:
		ax.legend()
	elif show_legend:
		ax.legend()

	fig.tight_layout(pad=0.5)
	result = plot_as_svg_xhtml(fig, **kwargs)
	plt.close(fig)

	return result





import numpy

class NotSameShapeError(ValueError):
	pass


def _sqz(shape):
	if not isinstance(shape, tuple):
		shape = tuple(shape)
	while len(shape) and shape[-1] == 1:
		shape = shape[:-1]
	return shape

def _sqz_same(shape1, shape2, silent=False, comment=None):
	if _sqz(shape1) != _sqz(shape2):
		if silent:
			return False
		if comment is not None:
			raise NotSameShapeError(f"shapes not same: {shape1} vs {shape2}   # {comment}")
		raise NotSameShapeError(f"shapes not same: {shape1} vs {shape2}")
	return True


def _sqz_same_trailing_neg_ok(shape1, shape2, silent=False, comment=None):
	s1, s2 = _sqz(shape1), _sqz(shape2)
	if s1 != s2:
		# check for match on all but trailing dim
		if s1[:-1] == s2[:-1]:
			# check for a negative in the trailing dim
			if s1[-1] < 0 or s2[-1] < 0:
				return True
		if silent:
			return False
		if comment is not None:
			raise NotSameShapeError(f"shapes not same: {shape1} vs {shape2}   # {comment}")
		raise NotSameShapeError(f"shapes not same: {shape1} vs {shape2}")
	return True



def selector_len_for(selector, seqlen):
	if selector is None:
		return seqlen
	if isinstance(selector, slice):
		start, stop, step = selector.indices(seqlen)
		return max(0, (stop - start + (step - (1 if step > 0 else -1))) // step)
	if isinstance(selector, numpy.ndarray):
		if selector.dtype == numpy.bool:
			if selector.shape[0] != seqlen:
				raise ValueError('bool array selector must be same size as unselected first dimension')
			return selector.sum()
		if numpy.issubdtype(selector.dtype, numpy.integer):
			return len(selector)
	raise TypeError('selector must be None or slice or ndarray')



def bitmask_shift_value(b):
	s = 1
	r = bin(b)
	while r[-s] == '0':
		s += 1
	return s-1


class SystematicAlternatives:

	__slots__ = ['masks', 'groupby', 'categoricals', 'altcodes', 'padding_levels']

	def __init__(self, masks, groupby, categoricals, altcodes, padding_levels):
		self.masks = masks
		self.groupby = groupby
		self.categoricals = categoricals
		self.altcodes = altcodes
		self.padding_levels = padding_levels

	def __repr__(self):
		s = "<larch.SystematicAlternatives>"
		for g,c in zip(self.groupby, self.categoricals):
			s += f"\n | {g}:"
			s += f"\n |   {str(c)}"
		return s

	def alternative_name_from_code(self, code):
		"""

		Parameters
		----------
		code : int
			The code integer for an alternative.

		Returns
		-------
		str
			A descriptive name for the alternative, made by concatenating the various
			categorical names with the within-category integer.

		"""
		name = ""
		for i,mask in enumerate(self.masks[:-1]):
			j = (mask & code) >> bitmask_shift_value(mask)
			j -= 1
			if j<0:
				name += "§-"
			else:
				try:
					name += f"{self.categoricals[i][j]}-"
				except IndexError:
					name += f"+{j+1-len(self.categoricals[i])}-"
		mask = self.masks[-1]
		j = (mask & code) >> bitmask_shift_value(mask)
		if j==0:
			name += f"§"
		else:
			name += f"{j}"
		return name

	@property
	def altnames(self):
		return [self.alternative_name_from_code(code) for code in self.altcodes]